from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator
from airflow.utils.context import Context
from kubernetes.client import models as k8s
from kubernetes import client, config
from kubernetes.stream import stream as kubernetes_stream
from typing import Dict, Optional, Any
from contextlib import closing
import ast
import copy

from starflow.utils import dict_deep_update
from starflow.client.starflow_backend_client import starflowBackendRestClient
from starflow.schemas import WorkflowSharedStorage, ContainerResourcesModel
from starflow.storage.s3 import S3StorageRepository
from starflow.logger import get_configured_logger
from airflow.exceptions import AirflowException
from airflow.providers.cncf.kubernetes.utils.xcom_sidecar import PodDefaults
import json
import os

class starflowKubernetesPodOperator(KubernetesPodOperator):
    def __init__(
        self,
        dag_id: str,
        task_id: str,
        piece_name: str,
        deploy_mode: str, # TODO enum
        repository_url: str,
        repository_version: str,
        workspace_id: int,
        username: Optional[str] = None,
        piece_input_kwargs: Optional[Dict] = None,
        workflow_shared_storage: WorkflowSharedStorage = None,
        container_resources: Optional[Dict] = None,
        **k8s_operator_kwargs
    ):
        import logging
        logger = logging.getLogger(__name__)
        
        logger.info("=" * 80)
        logger.info(f"[StarflowK8sOperator] INIT METHOD CALLED - task_id: {task_id}")
        logger.info(f"[StarflowK8sOperator] dag_id: {dag_id}")
        logger.info(f"[StarflowK8sOperator] piece_name: {piece_name}")
        logger.info(f"[StarflowK8sOperator] repository_url: {repository_url}")
        logger.info(f"[StarflowK8sOperator] repository_version: {repository_version}")
        logger.info(f"[StarflowK8sOperator] workspace_id: {workspace_id}")
        logger.info("=" * 80)
        
        self.logger = get_configured_logger("starflowKubernetesPodOperator")
        self.task_id = task_id
        self.piece_name = piece_name
        self.deploy_mode = deploy_mode
        self.repository_url = repository_url
        self.repository_version = repository_version
        self.workspace_id = workspace_id
        self.piece_input_kwargs = piece_input_kwargs
        self.workflow_shared_storage = workflow_shared_storage
        self.piece_source_image = k8s_operator_kwargs["image"]
        self.username_var = username
        
        # Initialize attributes needed by build_pod_request_obj
        self.task_id_replaced = self.task_id.lower().replace("_", "-") # doing this because airflow doesn't allow underscores and upper case in mount names and max len is 63
        self.shared_storage_base_mount_path = '/home/shared_storage'

        # Environment variables
        pod_env_vars = {
            "STARFLOW_PIECE": piece_name,
            "WORKSPACE_ID": str(workspace_id),
            "STARFLOW_INSTANTIATE_PIECE_KWARGS": str({
                "deploy_mode": deploy_mode,
                "task_id": task_id,
                "dag_id": dag_id,
            }),
            "STARFLOW_RUN_PIECE_KWARGS": str(piece_input_kwargs),
            "STARFLOW_WORKFLOW_SHARED_STORAGE": workflow_shared_storage.model_dump_json() if workflow_shared_storage else "",
            "AIRFLOW_CONTEXT_EXECUTION_DATETIME": "{{ dag_run.logical_date | ts_nodash }}",
            "AIRFLOW_CONTEXT_DAG_RUN_ID": "{{ run_id }}",
            # "OPEN_ID_SERVER_URL": os.environ['OPEN_ID_SERVER_URL'],
            # "OPEN_ID_CLIENT_ID": os.environ['OPEN_ID_CLIENT_ID'],
            # "OPEN_ID_REALM_NAME": os.environ['OPEN_ID_REALM_NAME'],
            # "OPEN_ID_CLIENT_SECRET": os.environ['OPEN_ID_CLIENT_SECRET'],
            # "URL_PUT_START_DOCK": os.environ['URL_PUT_START_DOCK'],
            # "URL_GET_FLEET_MOVEMENT_CALCULATION": os.environ['URL_GET_FLEET_MOVEMENT_CALCULATION'],
            # "URL_GET_FLEET_MINING_CALCULATION": os.environ['URL_GET_FLEET_MINING_CALCULATION'],
            # "URL_PUT_LOAD_CARGO": os.environ['URL_PUT_LOAD_CARGO'],
            # "URL_PUT_LOAD_AMMO": os.environ['URL_PUT_LOAD_AMMO'],
            # "URL_PUT_LOAD_FUEL": os.environ['URL_PUT_LOAD_FUEL'],
            # "URL_PUT_UNLOAD_CARGO": os.environ['URL_PUT_UNLOAD_CARGO'],
            # "URL_PUT_UNLOAD_AMMO": os.environ['URL_PUT_UNLOAD_AMMO'],
            # "URL_PUT_UNLOAD_FUEL": os.environ['URL_PUT_UNLOAD_FUEL'],
            # "URL_PUT_START_MINING": os.environ['URL_PUT_START_MINING'],
            # "URL_PUT_STOP_MINING": os.environ['URL_PUT_STOP_MINING'],
            # "URL_PUT_START_DOCK": os.environ['URL_PUT_START_DOCK'],
            # "URL_PUT_START_UNDOCK": os.environ['URL_PUT_START_UNDOCK'],
            # "URL_PUT_START_SUBWARP": os.environ['URL_PUT_START_SUBWARP'],
            # "URL_PUT_START_EXITSUBWARP": os.environ['URL_PUT_START_EXITSUBWARP'],
            # "URL_PUT_START_WARP": os.environ['URL_PUT_START_WARP'],
            # "URL_PUT_START_EXITWARP": os.environ['URL_PUT_START_EXITWARP'],
            # "URL_GET_LIST_FLEET": os.environ['URL_GET_LIST_FLEET'],
            # "OPEN_ID_USERNAME_SERVICE_USER": os.environ['OPEN_ID_USERNAME_SERVICE_USER'],
            # "OPEN_ID_PASSWORD_SERVICE_USER": os.environ['OPEN_ID_PASSWORD_SERVICE_USER']
        }

        # Container resources
        if container_resources is None:
            container_resources = {}
        base_container_resources_model = ContainerResourcesModel(
            requests={"cpu": "100m", "memory": "128Mi",},
            limits={"cpu": "100m", "memory": "128Mi"},
            use_gpu=False,
        )
        basic_container_resources = base_container_resources_model.model_dump()
        updated_container_resources = dict_deep_update(basic_container_resources, container_resources)
        use_gpu = updated_container_resources.pop("use_gpu", False)
        if use_gpu:
            updated_container_resources["limits"]["nvidia.com/gpu"] = "1"
        container_resources_obj = k8s.V1ResourceRequirements(**updated_container_resources)

        # Extra volume and volume mounts - for DEV mode only
        # volumes_dev, volume_mounts_dev = None, None
        # if self.deploy_mode == 'local-k8s-dev':
        #     volumes_dev, volume_mounts_dev = self._make_volumes_and_volume_mounts_dev()

        super().__init__(
            task_id=task_id,
            env_vars=pod_env_vars,
            container_resources=container_resources_obj,
            # volumes=volumes_dev,
            # volume_mounts=volume_mounts_dev,
            **k8s_operator_kwargs
        )


    def _make_volumes_and_volume_mounts_dev(self):
        """
        Make volumes and volume mounts for the pod when in DEVELOPMENT mode.
        """
        config.load_incluster_config()
        k8s_client = client.CoreV1Api()

        all_volumes = []
        all_volume_mounts = []

        repository_raw_project_name = str(self.piece_source_image).split('/')[-1].split(':')[0]
        persistent_volume_claim_name = 'pvc-{}'.format(str(repository_raw_project_name.lower().replace('_', '-')))
        persistent_volume_name = 'pv-{}'.format(str(repository_raw_project_name.lower().replace('_', '-')))

        pvc_exists = False
        try:
            k8s_client.read_namespaced_persistent_volume_claim(name=persistent_volume_claim_name, namespace='default')
            pvc_exists = True
        except client.rest.ApiException as e:
            if e.status != 404:
                raise e

        pv_exists = False
        try:
            k8s_client.read_persistent_volume(name=persistent_volume_name)
            pv_exists = True
        except client.rest.ApiException as e:
            if e.status != 404:
                raise e

        if pv_exists and pvc_exists:
            volume_dev_pieces = k8s.V1Volume(
                name='dev-op-{path_name}'.format(path_name=str(repository_raw_project_name.lower().replace('_', '-'))),
                persistent_volume_claim=k8s.V1PersistentVolumeClaimVolumeSource(
                    claim_name=persistent_volume_claim_name
                ),
            )
            volume_mount_dev_pieces = k8s.V1VolumeMount(
                name='dev-op-{path_name}'.format(path_name=str(repository_raw_project_name.lower().replace('_', '-'))),
                mount_path=f'/home/starflow/pieces_repository',
                sub_path=None,
                read_only=True
            )
            all_volumes.append(volume_dev_pieces)
            all_volume_mounts.append(volume_mount_dev_pieces)

        ######################## For local starflow-py dev ###############################################
        starflow_package_local_claim_name = 'starflow-dev-volume-claim'
        pvc_exists = False
        try:
            k8s_client.read_namespaced_persistent_volume_claim(name=starflow_package_local_claim_name, namespace='default')
            pvc_exists = True
        except client.rest.ApiException as e:
            if e.status != 404:
                raise e

        if pvc_exists:
            volume_dev = k8s.V1Volume(
                name='jobs-persistent-storage-dev',
                persistent_volume_claim=k8s.V1PersistentVolumeClaimVolumeSource(claim_name=starflow_package_local_claim_name),
            )
            """
            # TODO
            Remove deprecated_volume_mount_dev once we have all the pieces repositories updated
            with the new base pod image
            """
            volume_mount_pkg = k8s.V1VolumeMount(
                name='jobs-persistent-storage-dev',
                mount_path='/usr/local/lib/python3.10/site-packages/starflow/',
                sub_path=None,
                read_only=True
            )
            all_volumes.append(volume_dev)
            all_volume_mounts.append(volume_mount_pkg)

        return all_volumes, all_volume_mounts

    def build_pod_request_obj(self, context: Optional['Context'] = None) -> k8s.V1Pod:
        """
        We override this method to add the shared storage to the pod.
        This function is called by the parent class's execute method.
        """
        pod = super().build_pod_request_obj(context)

        # Fix pod name to comply with Kubernetes naming rules (no underscores)
        if pod.metadata.name:
            # Replace underscores with hyphens and ensure it's lowercase
            pod.metadata.name = pod.metadata.name.lower().replace("_", "-")
            # Ensure it starts and ends with alphanumeric characters
            pod.metadata.name = pod.metadata.name.strip("-")
            if not pod.metadata.name[0].isalnum():
                pod.metadata.name = "pod-" + pod.metadata.name
            if not pod.metadata.name[-1].isalnum():
                pod.metadata.name = pod.metadata.name + "-pod"

        pod.metadata.namespace = "airflow"
        pod.spec.image_pull_secrets = [k8s.V1LocalObjectReference('ghcr-login-secret')]

        # Override xcom sidecar image to use Harbor public registry instead of quay.io
        if pod.spec.containers:
            for container in pod.spec.containers:
                if container.name == "airflow-xcom-sidecar":
                    container.image = "harbor.prochain.name/public/alpine:latest"
                    self.logger.info(f"[StarflowK8sOperator] Updated xcom sidecar image to: {container.image}")

        # if not self.workflow_shared_storage or self.workflow_shared_storage.mode.name == 'none':
        #     return pod
        # if self.workflow_shared_storage.source.name in ["aws_s3", "gcs"]:
        #     pod = self.add_shared_storage_sidecar(pod)
        # elif self.workflow_shared_storage.source.name == "local":
        #     pod = self.add_local_shared_storage_volumes(pod)

        return pod

    def add_local_shared_storage_volumes(self, pod: k8s.V1Pod) -> k8s.V1Pod:
        """Adds local shared storage volumes to the pod."""
        pod_cp = copy.deepcopy(pod)
        pod_cp.spec.volumes = pod.spec.volumes or []
        pod_cp.spec.volumes.append(
            k8s.V1Volume(
                name=f'workflow-shared-storage-volume-{self.task_id_replaced}'[0:63], # max resource name in k8s is 63 chars
                persistent_volume_claim=k8s.V1PersistentVolumeClaimVolumeSource(claim_name="starflow-workflow-shared-storage-volume-claim")
            )
        )
        # Add volume mounts for upstream tasks, using subpaths
        pod_cp.spec.containers[0].volume_mounts = pod_cp.spec.containers[0].volume_mounts or []
        for tid in self.shared_storage_upstream_ids_list:
            pod_cp.spec.containers[0].volume_mounts.append(
                k8s.V1VolumeMount(
                    name=f'workflow-shared-storage-volume-{self.task_id_replaced}'[0:63], # max resource name in k8s is 63 chars
                    mount_path=f"{self.shared_storage_base_mount_path}/{tid}",  # path inside main container
                    mount_propagation="HostToContainer",
                    read_only=True,
                )
            )
        # Add volume mount for this task
        pod_cp.spec.containers[0].volume_mounts.append(
            k8s.V1VolumeMount(
                name=f'workflow-shared-storage-volume-{self.task_id_replaced}'[0:63],  # max resource name in k8s is 63 chars
                mount_path=f"{self.shared_storage_base_mount_path}/{self.task_id}",  # path inside main container
                mount_propagation="Bidirectional",
                read_only=True,
            )
        )
        return pod_cp

    def _validate_storage_piece_secrets(self, storage_piece_secrets: Dict[str, Any]):
        validated = False
        if self.workflow_shared_storage.source.name == 'aws_s3':
            s3_storage_repository = S3StorageRepository()
            validated = s3_storage_repository.validate_s3_credentials_access(
                access_key=storage_piece_secrets.get('AWS_ACCESS_KEY_ID'),
                secret_key=storage_piece_secrets.get('AWS_SECRET_ACCESS_KEY'),
                bucket=self.workflow_shared_storage.bucket,
            )
        return validated

    def add_shared_storage_sidecar(self, pod: k8s.V1Pod) -> k8s.V1Pod:
        """
        - add shared storage sidecar container to the pod
        - add shared storage volume mounts to the sidecar container
        - add FUSE mounts configuration as env vars to the sidecar container
        """
        # Set up pod Volumes and containers VolumemMounts from Upstream tasks
        volume_mounts_main_container = list()
        volume_mounts_sidecar_container = list()
        pod_volumes_list = list()
        for tid in self.shared_storage_upstream_ids_list:
            tid_replaced = tid.lower().replace("_", "-") # Doing this because k8s doesn't allow underscores in volume names
            volume_mounts_main_container.append(
                k8s.V1VolumeMount(
                    name=f'workflow-shared-storage-volume-{tid_replaced}'[0:63], # max resource name in k8s is 63 chars
                    mount_path=f"{self.shared_storage_base_mount_path}/{tid}",  # path inside main container
                    mount_propagation="HostToContainer",
                    read_only=True,
                )
            )
            volume_mounts_sidecar_container.append(
                k8s.V1VolumeMount(
                    name=f'workflow-shared-storage-volume-{tid_replaced}'[0:63], # max resource name in k8s is 63 chars
                    mount_path=f"{self.shared_storage_base_mount_path}/{tid}",  # path inside sidecar container
                    mount_propagation="Bidirectional",
                )
            )
            pod_volumes_list.append(
                k8s.V1Volume(
                    name=f'workflow-shared-storage-volume-{tid_replaced}'[0:63], # max resource name in k8s is 63 chars
                    empty_dir=k8s.V1EmptyDirVolumeSource()
                )
            )
        # Set up pod Volumes and containers VolumemMounts for this Operator results
        volume_mounts_main_container.append(
            k8s.V1VolumeMount(
                name=f'workflow-shared-storage-volume-{self.task_id_replaced}'[0:63], # max resource name in k8s is 63 chars
                mount_path=f"{self.shared_storage_base_mount_path}/{self.task_id}",  # path inside main container
                mount_propagation="HostToContainer",
                read_only=False,
            )
        )
        volume_mounts_sidecar_container.append(
            k8s.V1VolumeMount(
                name=f'workflow-shared-storage-volume-{self.task_id_replaced}'[0:63], # max resource name in k8s is 63 chars
                mount_path=f"{self.shared_storage_base_mount_path}/{self.task_id}",  # path inside sidecar container
                mount_propagation="Bidirectional",
            )
        )
        pod_volumes_list.append(
            k8s.V1Volume(
                name=f'workflow-shared-storage-volume-{self.task_id_replaced}'[0:63], # max resource name in k8s is 63 chars
                empty_dir=k8s.V1EmptyDirVolumeSource()
            )
        )
        pod_cp = copy.deepcopy(pod)
        pod_cp.spec.volumes = pod.spec.volumes or []
        pod_cp.spec.volumes.extend(pod_volumes_list)
        pod_cp.spec.containers[0].volume_mounts = pod_cp.spec.containers[0].volume_mounts or []
        pod_cp.spec.containers[0].volume_mounts.extend(volume_mounts_main_container)

        # Create and add sidecar container to pod
        storage_piece_secrets = {}
        if self.workflow_shared_storage.source != "local":
            storage_piece_secrets = self._get_piece_secrets(
                repository_url="starflow-default/default_storage_repository",
                repository_version="0.0.1",
                piece_name=self.workflow_shared_storage.storage_piece_name,
                source='default',
            )
        if not self._validate_storage_piece_secrets(storage_piece_secrets):
            self.logger.error("Invalid storage piece secrets. Aborting pod creation.")
            raise Exception("Invalid storage piece secrets. Aborting pod creation.")

        self.workflow_shared_storage.source = self.workflow_shared_storage.source.name
        sidecar_env_vars = {
            'STARFLOW_WORKFLOW_SHARED_STORAGE': self.workflow_shared_storage.model_dump_json() if self.workflow_shared_storage else "",
            'STARFLOW_WORKFLOW_SHARED_STORAGE_SECRETS': str(storage_piece_secrets),
            'STARFLOW_INSTANTIATE_PIECE_KWARGS': str({
                "deploy_mode": self.deploy_mode,
                "task_id": self.task_id,
                "dag_id": self.dag_id,
            }),
            'STARFLOW_WORKFLOW_RUN_SUBPATH': self.workflow_run_subpath,
            'AIRFLOW_UPSTREAM_TASKS_IDS_SHARED_STORAGE': str(self.shared_storage_upstream_ids_list),
        }
        self.shared_storage_sidecar_container_name = f"starflow-shared-storage-sidecar-{self.task_id_replaced}"[0:63]
        sidecar_container = k8s.V1Container(
            name=self.shared_storage_sidecar_container_name,
            command=['bash', '-c', './sidecar_lifecycle.sh'],
            image='ghcr.io/prochain-star-atlas/starflow-shared-storage-sidecar:latest',
            volume_mounts=volume_mounts_sidecar_container,
            security_context=k8s.V1SecurityContext(privileged=True),
            env=[k8s.V1EnvVar(name=k, value=v) for k, v in sidecar_env_vars.items()],
            resources=k8s.V1ResourceRequirements(
                requests={
                    "cpu": "1m",
                    "memory": "128Mi",
                },
                limits={
                    "cpu": "1000m",
                    "memory": "1024Mi",
                }
            ),
        )
        pod_cp.spec.containers.append(sidecar_container)
        return pod_cp

    def _get_piece_secrets(
        self,
        repository_url: str,
        repository_version: str,
        piece_name: str,
        source: str = 'github'
    ) -> Dict[str, Any]:
        """Get piece secrets values from starflow API"""
        import logging
        logger = logging.getLogger(__name__)
        
        try:
            logger.info(f"[StarflowK8sOperator] Getting piece secrets for piece: {piece_name}")
            logger.info(f"[StarflowK8sOperator] Repository URL: {repository_url}")
            logger.info(f"[StarflowK8sOperator] Repository version: {repository_version}")
            logger.info(f"[StarflowK8sOperator] Workspace ID: {self.workspace_id}")
            
            params = {
                "workspace_id": self.workspace_id,
                "url": repository_url,
                "version": repository_version,
                'source': source,
                "page": 0,
                "page_size": 1,
            }
            
            logger.info("[StarflowK8sOperator] Calling get_piece_repositories_from_workspace_id...")
            piece_repository_response = self.starflow_client.get_piece_repositories_from_workspace_id(
                params=params
            )
            logger.info(f"[StarflowK8sOperator] get_piece_repositories_from_workspace_id response status: {piece_repository_response.status_code}")
            
            piece_repository_data = piece_repository_response.json()
            logger.info(f"[StarflowK8sOperator] Piece repository data: {piece_repository_data}")
            
            # Check if any piece repositories were found
            if not piece_repository_data.get("data") or len(piece_repository_data["data"]) == 0:
                logger.warning(f"[StarflowK8sOperator] No piece repositories found for workspace_id={self.workspace_id}, url={repository_url}, version={repository_version}")
                logger.info("[StarflowK8sOperator] Returning empty secrets dict")
                return {}
            
            piece_repository_id = piece_repository_data["data"][0]["id"]
            logger.info(f"[StarflowK8sOperator] Piece repository ID: {piece_repository_id}")
            
            logger.info("[StarflowK8sOperator] Calling get_piece_secrets...")
            secrets_response = self.starflow_client.get_piece_secrets(
                piece_repository_id=piece_repository_id,
                piece_name=piece_name
            )
            logger.info(f"[StarflowK8sOperator] get_piece_secrets response status: {secrets_response.status_code}")
            
            if secrets_response.status_code == 404:
                logger.info("[StarflowK8sOperator] No secrets configured for this piece, returning empty dict")
                return {}
            elif secrets_response.status_code != 200:
                logger.error(f"[StarflowK8sOperator] Error getting piece secrets: {secrets_response.json()}")
                raise Exception(f"Error getting piece secrets: {secrets_response.json()}")

            secrets_data = secrets_response.json()
            logger.info(f"[StarflowK8sOperator] Secrets data: {secrets_data}")
            
            piece_secrets = {}
            for e in secrets_data:
                if not e.get('value') and not e.get('required'):
                    continue
                piece_secrets[e.get('name')] = e.get('value')
            
            logger.info(f"[StarflowK8sOperator] Piece secrets processed: {list(piece_secrets.keys())}")
            return piece_secrets
            
        except Exception as e:
            logger.error(f"[StarflowK8sOperator] Error in _get_piece_secrets: {str(e)}")
            logger.error(f"[StarflowK8sOperator] Exception type: {type(e).__name__}")
            import traceback
            logger.error(f"[StarflowK8sOperator] Traceback: {traceback.format_exc()}")
            raise


    @staticmethod
    def _get_upstream_xcom_data_from_task_ids(task_ids: list, context: Context):
        upstream_xcoms_data = dict()
        for tid in task_ids:
            upstream_xcoms_data[tid] = context['ti'].xcom_pull(task_ids=tid)
        return upstream_xcoms_data

    def _get_piece_kwargs_value_from_upstream_xcom(
        self,
        value: Any
    ):
        if isinstance(value, dict) and value.get("type") == "fromUpstream":
            upstream_task_id = value["upstream_task_id"]
            output_arg = value["output_arg"]
            if upstream_task_id not in self.shared_storage_upstream_ids_list:
                self.shared_storage_upstream_ids_list.append(upstream_task_id)
            return self.upstream_xcoms_data[upstream_task_id][output_arg]
        elif isinstance(value, list):
            return [self._get_piece_kwargs_value_from_upstream_xcom(item) for item in value]
        elif isinstance(value, dict):
            return {
                k: self._get_piece_kwargs_value_from_upstream_xcom(v)
                for k, v in value.items()
            }
        return value

    def _update_piece_kwargs_with_upstream_xcom(self):
        self.shared_storage_upstream_ids_list = list()
        if not self.piece_input_kwargs:
            self.piece_input_kwargs = dict()

        updated_piece_kwargs = {
            k: self._get_piece_kwargs_value_from_upstream_xcom(
                value=v
            ) for k, v in self.piece_input_kwargs.items()
        }
        self.piece_input_kwargs = updated_piece_kwargs
        self.env_vars.append({
            'name': 'AIRFLOW_UPSTREAM_TASKS_IDS_SHARED_STORAGE',
            'value': str(self.shared_storage_upstream_ids_list),
            'value_from': None
        })

    def _update_env_var_value_from_name(self, name: str, value: str):
        for env_var in self.env_vars:
            if env_var.name == name:
                env_var.value = value
                break


    def _prepare_execute_environment(self, context: Context):
        """
        Runs at the begining of the execute method.
        Pass extra arguments and configuration as environment variables to the pod
        """
        import logging
        logger = logging.getLogger(__name__)
        
        try:
            logger.info("[StarflowK8sOperator] Fetching upstream task IDs...")
            # Fetch upstream tasks ids and save them in an ENV var
            # Check if operator is properly assigned to a DAG
            try:
                # Try to access the dag property - this will raise RuntimeError if not assigned
                dag = self.dag
                if dag is None:
                    logger.warning("[StarflowK8sOperator] Operator DAG is None, skipping upstream task IDs")
                    upstream_task_ids = []
                else:
                    upstream_task_ids = [t.task_id for t in self.get_direct_relatives(upstream=True)]
                    logger.info(f"[StarflowK8sOperator] Upstream task IDs: {upstream_task_ids}")
            except RuntimeError as e:
                if "has not been assigned to a DAG yet" in str(e):
                    logger.warning("[StarflowK8sOperator] Operator not assigned to a DAG, skipping upstream task IDs")
                    upstream_task_ids = []
                else:
                    raise
            
            self.env_vars.append({
                'name': 'AIRFLOW_UPSTREAM_TASKS_IDS',
                'value': str(upstream_task_ids),
                'value_from': None
            })
            
            # Pass forward the workflow shared storage source name
            logger.info("[StarflowK8sOperator] Setting workflow shared storage environment variable...")
            self.env_vars.append({
                'name': 'STARFLOW_WORKFLOW_SHARED_STORAGE_SOURCE_NAME',
                'value': str(self.workflow_shared_storage.source.name) if self.workflow_shared_storage else None,
                'value_from': None
            })
            
            # Save updated piece input kwargs with upstream data to environment variable
            logger.info("[StarflowK8sOperator] Getting upstream XCom data...")
            self.upstream_xcoms_data = self._get_upstream_xcom_data_from_task_ids(task_ids=upstream_task_ids, context=context)
            logger.info("[StarflowK8sOperator] Upstream XCom data retrieved successfully")
            
            logger.info("[StarflowK8sOperator] Updating piece kwargs with upstream XCom...")
            self._update_piece_kwargs_with_upstream_xcom()
            logger.info("[StarflowK8sOperator] Piece kwargs updated successfully")
            
            logger.info("[StarflowK8sOperator] Getting piece secrets...")
            piece_secrets = self._get_piece_secrets(
                repository_url=self.repository_url,
                repository_version=self.repository_version,
                piece_name=self.piece_name,
                source='github'
            )
            logger.info(f"[StarflowK8sOperator] Piece secrets retrieved: {list(piece_secrets.keys())}")
            
            logger.info("[StarflowK8sOperator] Setting piece secrets environment variables...")
            for secret_name, secret_value in piece_secrets.items():
                self.env_vars.append({
                    'name': secret_name,
                    'value': str(secret_value),
                    'value_from': None
                })
            logger.info("[StarflowK8sOperator] Piece secrets environment variables set successfully")
            
        except Exception as e:
            logger.error(f"[StarflowK8sOperator] Error in _prepare_execute_environment: {str(e)}")
            logger.error(f"[StarflowK8sOperator] Exception type: {type(e).__name__}")
            import traceback
            logger.error(f"[StarflowK8sOperator] Traceback: {traceback.format_exc()}")
            raise
        
        logger.info("[StarflowK8sOperator] Updating STARFLOW_RUN_PIECE_KWARGS environment variable...")
        self._update_env_var_value_from_name(name='STARFLOW_RUN_PIECE_KWARGS', value=str(self.piece_input_kwargs))
        logger.info("[StarflowK8sOperator] STARFLOW_RUN_PIECE_KWARGS environment variable updated successfully")
        # self.env_vars.append({
        #     "name": "STARFLOW_PIECE_SECRETS",
        #     "value": str(piece_secrets),
        #     "value_from": None
        # })

        # Include workflow run subpath from dag run id, taken from context
        dag_id = context["dag_run"].dag_id
        dag_run_id = context['run_id']
        dag_run_id_path = dag_run_id.replace("-", "_").replace(".", "_").replace(" ", "_").replace(":", "_").replace("+", "_")
        self.workflow_run_subpath = f"{dag_id}/{dag_run_id_path}"
        self.env_vars.append({
            'name': 'STARFLOW_WORKFLOW_RUN_SUBPATH',
            'value': self.workflow_run_subpath,
            'value_from': None
        })
        self.env_vars.append({
            'name': 'OPEN_ID_USERNAME_TARGET',
            'value': str(self.username_var),
            'value_from': None
        })


    def execute(self, context: Context):
        """
        Code from here onward is executed by the Worker and not by the Scheduler.
        """
        import logging
        logger = logging.getLogger(__name__)
        
        # Log at the very beginning to confirm execute method is called
        logger.info("=" * 80)
        logger.info(f"[StarflowK8sOperator] EXECUTE METHOD CALLED - task_id: {self.task_id}")
        logger.info(f"[StarflowK8sOperator] Context type: {type(context)}")
        logger.info(f"[StarflowK8sOperator] Context keys: {list(context.keys()) if hasattr(context, 'keys') else 'No keys method'}")
        logger.info("=" * 80)
        
        try:
            logger.info(f"[StarflowK8sOperator] Starting execution for task_id: {self.task_id}")
            
            # Initialize remote_pod early to avoid UnboundLocalError in finally block
            remote_pod = None
            
            # TODO change url based on platform configuration
            logger.info("[StarflowK8sOperator] Initializing Starflow backend client...")
            self.starflow_client = starflowBackendRestClient(base_url="http://starflow-rest-service.prochain-starflow:8000/")
            logger.info("[StarflowK8sOperator] Starflow backend client initialized successfully")
            
            logger.info("[StarflowK8sOperator] Preparing execute environment...")
            self._prepare_execute_environment(context=context)
            logger.info("[StarflowK8sOperator] Execute environment prepared successfully")
            logger.info("[StarflowK8sOperator] Building pod request object...")
            self.pod_request_obj = self.build_pod_request_obj(context)
            logger.info("[StarflowK8sOperator] Pod request object built successfully")
            
            logger.info("[StarflowK8sOperator] Creating/getting pod...")
            self.pod = self.get_or_create_pod(  # must set `self.pod` for `on_kill`
                pod_request_obj=self.pod_request_obj,
                context=context,
            )
            logger.info(f"[StarflowK8sOperator] Pod created successfully: {self.pod.metadata.name}")
            
            # get remote pod for use in cleanup methods
            logger.info("[StarflowK8sOperator] Finding remote pod...")
            remote_pod = self.find_pod(self.pod.metadata.namespace, context=context)
            logger.info("[StarflowK8sOperator] Remote pod found successfully")
            
            logger.info("[StarflowK8sOperator] Awaiting pod start...")
            self.await_pod_start(pod=self.pod)
            logger.info("[StarflowK8sOperator] Pod started successfully")

            if self.get_logs:
                self.pod_manager.fetch_container_logs(
                    pod=self.pod,
                    container_name=self.BASE_CONTAINER_NAME,
                    follow=True,
                )
            else:
                self.pod_manager.await_container_completion(
                    pod=self.pod, container_name=self.BASE_CONTAINER_NAME
                )

            if self.do_xcom_push:
                result = self.extract_xcom(pod=self.pod)

            if self.workflow_shared_storage and self.workflow_shared_storage.mode.name != 'none':
                self._kill_shared_storage_sidecar(pod=self.pod)
            remote_pod = self.pod_manager.await_pod_completion(self.pod)
        finally:
            self.cleanup(
                pod=self.pod or self.pod_request_obj,
                remote_pod=remote_pod,
            )
        ti = context['ti']
        ti.xcom_push(key='pod_name', value=self.pod.metadata.name)
        ti.xcom_push(key='pod_namespace', value=self.pod.metadata.namespace)
        if self.do_xcom_push:
            self._shared_storage_usage_in_bytes = result.get('_shared_storage_usage_in_bytes', 0)
            return result


    def _kill_shared_storage_sidecar(self, pod: k8s.V1Pod):
        """
        This method is used to send a signal to stop and delete the sidecar container with the shared storage mounts.
        """
        with closing(
            kubernetes_stream(
                self.pod_manager._client.connect_get_namespaced_pod_exec,
                pod.metadata.name,
                pod.metadata.namespace,
                container=self.shared_storage_sidecar_container_name,
                command=['/bin/bash'],
                stdin=True,
                stdout=True,
                stderr=True,
                tty=False,
                _preload_content=False,
            )
        ) as resp:
            self.log.info('Sending signal to unmount shared storage sidecar container')
            self.pod_manager._exec_pod_command(resp, 'python mount.py unmount')
            self.log.info('Sending signal to delete shared storage sidecar container')
            self.pod_manager._exec_pod_command(resp, 'kill -s SIGINT 1')

    def extract_xcom(self, pod: k8s.V1Pod):
        """Retrieves xcom value and kills xcom sidecar container"""
        result = self.pod_manager_extract_xcom(pod)
        if isinstance(result, str) and result.rstrip() == "__airflow_xcom_result_empty__":
            self.log.info("Result file is empty.")
            return None
        else:
            self.log.info('xcom result is not empty, parsing to json')
            return json.loads(result)

    def pod_manager_extract_xcom(self, pod: k8s.V1Pod) -> str:
        client = kubernetes_stream(
            self.pod_manager._client.connect_get_namespaced_pod_exec,
            pod.metadata.name,
            pod.metadata.namespace,
            container=PodDefaults.SIDECAR_CONTAINER_NAME,
            command=[
                '/bin/sh',
                '-c',
                f"if [ -s {PodDefaults.XCOM_MOUNT_PATH}/return.json ]; then cat {PodDefaults.XCOM_MOUNT_PATH}/return.json; else echo __airflow_xcom_result_empty__; fi",
            ],
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=False,
            _request_timeout=10,
        )
        client.run_forever(timeout=10)
        result = client.read_all()

        _ = kubernetes_stream(
            self.pod_manager._client.connect_get_namespaced_pod_exec,
            pod.metadata.name,
            pod.metadata.namespace,
            container=PodDefaults.SIDECAR_CONTAINER_NAME,
            command=[
                '/bin/sh',
                '-c',
                'kill -s SIGINT 1',
            ],
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=True,
            _request_timeout=10,
        )
        client.close()

        if result is None:
            raise AirflowException(f"Failed to extract xcom from pod: {pod.metadata.name}")

        return result






from pydantic import BaseModel, Field
from enum import Enum


class OutputModifierItemType(str, Enum):
    """
    OutputArgsType Enum
    """
    string = 'string'
    integer = 'integer'
    float = 'float'
    boolean = 'boolean'
    array = 'array'


class OutputModifierModel(BaseModel):
    name: str = Field(
        description='Name of the output argument.',
        json_schema_extra={
            "from_upstream": "never"
        }
    )
    type: OutputModifierItemType = Field(
        default=OutputModifierItemType.string,
        description='Type of the output argument.',
        json_schema_extra={
            "from_upstream": "never"
        },
    )
    description: str = Field(
        default="",
        description='Description of the output argument.',
        json_schema_extra={
            "from_upstream": "never"
        }
    )



from starflow.scripts.load_piece import load_piece_class_from_path, load_piece_models_from_path
from pathlib import Path
import os
import ast
import json


def run_piece():
    # Import Operator from File System, already configured with metadata
    piece_dict = ast.literal_eval(os.getenv("STARFLOW_BASHOPERATOR_PIECE"))
    piece_name = piece_dict.get("name")

    pieces_repository_copy_path = Path("opt/airflow/starflow/pieces_repository")
    operators_folder_path = pieces_repository_copy_path / "pieces"
    compiled_metadata_path = pieces_repository_copy_path / ".starflow/compiled_metadata.json"
    
    with open(str(compiled_metadata_path), "r") as f:
        compiled_metadata = json.load(f)

    piece_class = load_piece_class_from_path(
        operators_folder_path=operators_folder_path,
        piece_name=piece_name,
        piece_metadata=compiled_metadata[piece_name]
    )

    piece_input_model_class, piece_output_model_class, piece_secrets_model_class = load_piece_models_from_path(
        pieces_folder_path=operators_folder_path,
        piece_name=piece_name
    )

    # Instantiate and run Operator
    instantiate_op_dict = ast.literal_eval(os.getenv("STARFLOW_BASHOPERATOR_INSTANTIATE_PIECE_KWARGS"))
    piece_object = piece_class(**instantiate_op_dict)

    run_op_dict = ast.literal_eval(os.getenv("STARFLOW_BASHOPERATOR_RUN_PIECE_KWARGS"))

    # Get relevant airflow context from ENV
    airflow_context = {
        "dag_owner": os.getenv('AIRFLOW_CONTEXT_DAG_OWNER'),
        "dag_id": os.getenv('AIRFLOW_CONTEXT_DAG_ID'),
        "task_id": os.getenv('AIRFLOW_CONTEXT_TASK_ID'),
        "execution_date": os.getenv('AIRFLOW_CONTEXT_EXECUTION_DATE'),
        "try_number": os.getenv('AIRFLOW_CONTEXT_TRY_NUMBER'),
        "dag_run_id": os.getenv('AIRFLOW_CONTEXT_DAG_RUN_ID'),
        "upstream_task_ids": os.getenv('AIRFLOW_CONTEXT_UPSTREAM_TASK_IDS')  # Added by starflow extended BashOperator
    }

    piece_object.run_piece_function(
        piece_input_model=piece_input_model_class, 
        piece_output_model=piece_output_model_class, 
        piece_secrets_model=piece_secrets_model_class,
        op_kwargs=run_op_dict,
        airflow_context=airflow_context
    )

    return None


import copy
import os
from starflow.scripts.docker_compose_scripts import init_script


# TODO - UPDATE THIS

# Airflow configuration variables reference:
# https://airflow.apache.org/docs/apache-airflow/stable/configurations-ref.html


# Variables shared by most services
code_repository_source = os.environ.get('CODE_REPOSITORY_SOURCE', 'local')
volume_mount_path_host = os.environ.get('VOLUME_MOUNT_PATH_HOST', './mnt/fs')
volume_mount_path_docker = "/opt/mnt/fs"
code_repository_path_host = os.environ.get('CODE_REPOSITORY_PATH_HOST', '.')
flowui_path_host = os.environ.get('FLOWUI_PATH_HOST', '../flowui')  # used for local dev only
flowui_path_docker = "/opt/flowui"  # used for local dev only

airflow_home_docker = "/opt/mnt/fs/airflow"
airflow_common_env = {
    "VOLUME_MOUNT_PATH_DOCKER": volume_mount_path_docker,
    "VOLUME_MOUNT_PATH_HOST": volume_mount_path_host,
    "AIRFLOW_UID": "${AIRFLOW_UID}",
    "AIRFLOW_HOME": airflow_home_docker,
    "AWS_ACCESS_KEY_ID": "${AWS_ACCESS_KEY_ID}",
    "AWS_SECRET_ACCESS_KEY": "${AWS_SECRET_ACCESS_KEY}",
    "AWS_REGION_NAME": "${AWS_REGION_NAME}",
    "AIRFLOW__CORE__DAGS_FOLDER": f"{airflow_home_docker}/dags",
    "AIRFLOW__CORE__PLUGINS_FOLDER": f"{airflow_home_docker}/plugins",
    "AIRFLOW__LOGGING__BASE_LOG_FOLDER": f"{airflow_home_docker}/logs",
    "AIRFLOW__LOGGING__DAG_PROCESSOR_MANAGER_LOG_LOCATION": f"{airflow_home_docker}/logs/dag_processor_manager/dag_processor_manager.log",
    "AIRFLOW__SCHEDULER__CHILD_PROCESS_LOG_DIRECTORY": f"{airflow_home_docker}/logs/scheduler",
    "AIRFLOW_WEBSERVER_HOST": "http://localhost:8080",
    # Default from original docker-compose.yaml
    "AIRFLOW__CORE__EXECUTOR": "CeleryExecutor",
    "AIRFLOW__CORE__SQL_ALCHEMY_CONN": "postgresql+psycopg2://airflow:airflow@postgres/airflow",
    "AIRFLOW__CELERY__RESULT_BACKEND": "db+postgresql://airflow:airflow@postgres/airflow",
    "AIRFLOW__CELERY__BROKER_URL": "redis://:@redis:6379/0",
    "AIRFLOW__CORE__FERNET_KEY": '',
    "AIRFLOW__CORE__DAGS_ARE_PAUSED_AT_CREATION": 'true',
    "AIRFLOW__CORE__LOAD_EXAMPLES": 'false',
    "AIRFLOW__CORE__ENABLE_XCOM_PICKLING": 'false',
    "AIRFLOW__API__AUTH_BACKEND": 'airflow.api.auth.backend.basic_auth',
    "_PIP_ADDITIONAL_REQUIREMENTS": "${_PIP_ADDITIONAL_REQUIREMENTS:-}",
    "AIRFLOW__SCHEDULER__DAG_DIR_LIST_INTERVAL": 60
}


airflow_common_volumes = [
    # "./mnt/fs:/opt/mnt/fs",
    f"{volume_mount_path_host}:{volume_mount_path_docker}",
]
if code_repository_source == 'local':
    airflow_common_volumes.append(
        f"{code_repository_path_host}:{volume_mount_path_docker}/code_repository"
    )

airflow_common_volumes_dev = [
    f"{volume_mount_path_host}:{volume_mount_path_docker}",
    f"{flowui_path_host}:{flowui_path_docker}",  # for hot-reload on dev
    f"{code_repository_path_host}:{volume_mount_path_docker}/code_repository"
]

airflow_common_depends_on = {
    "redis": {
        "condition": "service_healthy"
    },
    "postgres": {
        "condition": "service_healthy"
    }
}

# Airflow common
x_airflow_common = {
    "image": 'apache/airflow:2.3.3',
    "environment": airflow_common_env,
    "volumes": airflow_common_volumes,
    "user": "${AIRFLOW_UID:-50000}:0",
    "depends_on": airflow_common_depends_on
}

# Airflow common - for internal development
x_airflow_common_dev = {
    "image": 'apache/airflow:2.3.2',
    "environment": airflow_common_env,
    "volumes": airflow_common_volumes_dev,
    "user": "${AIRFLOW_UID:-50000}:0",
    "depends_on": airflow_common_depends_on
}

# Airflow webserver
airflow_webserver = {
    **x_airflow_common,
    "environment": airflow_common_env,
    "container_name": "airflow-webserver",
    "command": "webserver",
    "ports": [
        "8080:8080"
    ],
    "healthcheck": {
        "test":[
            "CMD",
            "curl",
            "--fail",
            "http://localhost:8080/health"
        ],
        "interval": "10s",
        "timeout": "10s",
        "retries": 5
    },
    "restart": "always",
    "depends_on": {
        **airflow_common_depends_on,
        "airflow-init": {
            "condition": "service_completed_successfully"
        }
    }
}

airflow_webserver_dev = copy.deepcopy(airflow_webserver)
airflow_webserver_dev.update({
    'volumes': airflow_common_volumes_dev,
})


# Airflow Redis service
airflow_redis = {
    "image": "redis:latest",
    "container_name": "airflow-redis",
    "expose": [
        6379
    ],
    "healthcheck": {
        "test": [
            "CMD",
            "redis-cli",
            "ping"
        ],
        "interval": "5s",
        "timeout": "30s",
        "retries": 50
    },
    "restart": "always"
}


# Airflow Postgres service
airflow_postgres = {
    "image": "postgres:14",
    "container_name": "postgres-airflow",
    "environment": {
        "POSTGRES_USER": "airflow",
        "POSTGRES_PASSWORD": "airflow",
        "POSTGRES_DB": "airflow"
    },
    "volumes": [
        "postgres-airflow-volume:/var/lib/postgresql/data"
    ],
    "healthcheck": {
        "test": [
            "CMD",
            "pg_isready",
            "-U",
            "airflow"
        ],
        "interval": "5s",
        "retries": 5
    },
    "restart": "always",
    "ports": [
        "5432:5432" # for local tests only
    ]
}


# Airflow Scheduler service
airflow_scheduler = {
    **x_airflow_common,
    "command": 'bash -c "pip install flowui && airflow scheduler"',
    "environment": airflow_common_env,
    "healthcheck": {
        "test": [
            "CMD-SHELL",
            'airflow jobs check --job-type SchedulerJob --hostname "$${HOSTNAME}"'
        ],
        "interval": "10s",
        "timeout": "10s",
        "retries": 5
    },
    "restart": "always",
    "depends_on": {
        **airflow_common_depends_on,
        "airflow-init": {
            "condition": "service_completed_successfully"
        }
    }
}


# Airflow Scheduler service - for internal development
airflow_scheduler_dev = copy.deepcopy(airflow_scheduler)
airflow_scheduler_dev.update({
    "command": f'bash -c "pip install -e {flowui_path_docker} && airflow scheduler"',  # for hot-reload on dev
    "volumes": airflow_common_volumes_dev
})


# Airflow Triggerer service
airflow_triggerer = {
    **x_airflow_common,
    "command": "triggerer",
    "environment": airflow_common_env,
    "healthcheck": {
        "test": [
            "CMD-SHELL",
            'airflow jobs check --job-type TriggererJob --hostname "$${HOSTNAME}"'
        ],
        "interval": "10s",
        "timeout": "10s",
        "retries": 5
    },
    "restart": "always",
    "depends_on": {
        **airflow_common_depends_on,
        "airflow-init": {
            "condition": "service_completed_successfully"
        }
    }
}


# Airflow CLI service
airflow_cli = {
    "profiles": [
        "debug"
    ],
    **x_airflow_common,
    "environment":{
        **airflow_common_env,
        "CONNECTION_CHECK_MAX_COUNT": "0"
    },
    "command": [
        "bash",
        "-c",
        "airflow"
    ]
}


# Airflow Init service
airflow_init = {
    **x_airflow_common,
    "container_name": "airflow-init",
    "entrypoint": "/bin/bash",
    "command": [
        "-c",
        f'{init_script}'
    ],
    "environment": {
        **airflow_common_env,
        "_AIRFLOW_DB_UPGRADE": 'true',
        "_AIRFLOW_WWW_USER_CREATE": 'true',
        "_AIRFLOW_WWW_USER_USERNAME": "${_AIRFLOW_WWW_USER_USERNAME:-airflow}",
        "_AIRFLOW_WWW_USER_PASSWORD": "${_AIRFLOW_WWW_USER_PASSWORD:-airflow}"
    },
    "user": "0:0",
    "volumes": [
        # f"{volume_mount_path_host}/airflow:/sources"
        f"{volume_mount_path_host}/airflow:{airflow_home_docker}"
    ] 
}


# Flower service
flower = {
    **x_airflow_common,
    "command": "celery flower",
    "ports": [
        "5555:5555"
    ],
    "healthcheck":{
        "test": [ "CMD", "curl", "--fail", "http://localhost:5555/" ],
        "interval": "10s",
        "timeout": "10s",
        "retries": 5
    },
    "restart": "always",
    "depends_on": {
        **airflow_common_depends_on,
        "airflow-init": {
            "condition": "service_completed_successfully"
        }
    }
}


# starflow backend service
flowui_backend = {
    "image": "atlassistant/flowui-backend:latest",
    "container_name": "flowui-backend",
    "command": 'bash -c "alembic upgrade heads && uvicorn main:app --reload --workers 1 --host 0.0.0.0 --port 8000"',
    "ports": [
        "8000:8000"
    ],
    "environment": [
        "STARFLOW_DB_USER=postgres",
        "STARFLOW_DB_PASSWORD=postgres",
        "STARFLOW_DB_HOST=flowui-postgres",
        "STARFLOW_DB_PORT=5432",
        "STARFLOW_DB_NAME=postgres",
        "STARFLOW_GITHUB_ACCESS_TOKEN_PIECES=${STARFLOW_GITHUB_ACCESS_TOKEN_PIECES}",
        "STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS=${STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS}",
        "STARFLOW_GITHUB_WORKFLOWS_REPOSITORY=${STARFLOW_GITHUB_WORKFLOWS_REPOSITORY}",
        "STARFLOW_DEPLOY_MODE=${STARFLOW_DEPLOY_MODE}"
    ],
    "networks": [
        "flowui-postgres-network"
    ],
    "volumes": airflow_common_volumes,
    "depends_on": {
        "flowui-postgres": {
            "condition": "service_healthy"
        }
    }
}

# starflow backend service - for internal development
flowui_backend_dev = {
    "build": {
        "context": f"{flowui_path_host}/backend",
        "dockerfile": "Dockerfile"
    },
    "container_name": "flowui-backend-dev",
    "command": 'bash -c "alembic upgrade heads && uvicorn main:app --reload --workers 1 --host 0.0.0.0 --port 8000"',
    "ports": [
        "8000:8000"
    ],
    "environment": [
        "STARFLOW_DB_USER=postgres",
        "STARFLOW_DB_PASSWORD=postgres",
        "STARFLOW_DB_HOST=flowui-postgres",
        "STARFLOW_DB_PORT=5432",
        "STARFLOW_DB_NAME=postgres",
        "STARFLOW_GITHUB_ACCESS_TOKEN_PIECES=${STARFLOW_GITHUB_ACCESS_TOKEN_PIECES}",
        "STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS=${STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS}",
        "STARFLOW_GITHUB_WORKFLOWS_REPOSITORY=${STARFLOW_GITHUB_WORKFLOWS_REPOSITORY}",
        "STARFLOW_DEPLOY_MODE=${STARFLOW_DEPLOY_MODE}"
    ],
    "networks": [
        "flowui-postgres-network"
    ],
    "volumes": airflow_common_volumes_dev + [f"{flowui_path_host}/backend:/backend"],
    "depends_on": {
        "flowui-postgres": {
            "condition": "service_healthy"
        }
    }
}


# starflow Postgres service
flowui_postgres = {
    "image": "postgres",
    "container_name": "flowui-postgres",
    "environment": [
        "POSTGRES_DB=postgres",
        "POSTGRES_USER=postgres",
        "POSTGRES_PASSWORD=postgres"
    ],
    "volumes": [
        "flowui-postgres-volume:/var/lib/postgresql/data"
    ],
    "networks": [
        "flowui-postgres-network"
    ],
    "healthcheck": {
        "test": [
            "CMD-SHELL",
            "pg_isready -d $${POSTGRES_DB} -U $${POSTGRES_USER}"
        ],
        "interval": "10s",
        "timeout": "60s",
        "retries": 5,
        "start_period": "2s"
    },
    "ports": [
        "5433:5432"
    ],
    "restart": "always"
}


# starflow Frontend service
flowui_frontend = {
    "image": "atlassistant/flowui-frontend:latest",
    "container_name": "flowui-frontend",
    "command": "pnpm start",
    "ports": [
        "3000:3000"
    ]
}

# starflow Frontend service - for internal development
flowui_frontend_dev = {
    "build": {
        "context": "${FLOWUI_PATH_HOST}/frontend",
        "dockerfile": "Dockerfile"
    },
    "container_name": "flowui-frontend",
    "command": "pnpm start",
    "ports": [
        "3000:3000"
    ],
    "environment": {
        "FLOWUI_PATH_HOST": "${FLOWUI_PATH_HOST:-../flowui}"
    },
    "volumes": [
        f"{flowui_path_host}/frontend:/frontend",  # For hot-reload on dev
    ]
}



from rich.console import Console
from pathlib import Path
import click
import os
import uuid
import tomli
from starflow.cli.utils import pieces_repository, platform
import ast
import starflow

console = Console()

# Ref: https://patorjk.com/software/taag/
# Font: standard
msg = """
==============================================
  ____                  _
 |  _ \  ___  _ __ ___ (_)_ __   ___
 | | | |/ _ \| '_ ` _ \| | '_ \ / _ \
 | |_| | (_) | | | | | | | | | | (_) |
 |____/ \___/|_| |_| |_|_|_| |_|\___/

=============================================="""


###############################################################################
# starflow PLATFORM
###############################################################################

def validate_github_token_workflows(value):
    if value and value.startswith("ghp_"):
        return value
    return None


def validate_github_token(value):
    if value and value.startswith("ghp_"):
        return value
    return None


def get_cluster_name_from_env():
    cluster_name = os.environ.get("STARFLOW_KIND_CLUSTER_NAME", None)
    if not cluster_name:
        cluster_name = "starflow-cluster"
    return cluster_name


def get_github_workflows_ssh_private_key_from_env():
    return os.environ.get("STARFLOW_GITHUB_WORKFLOWS_SSH_PRIVATE_KEY", "")


def get_github_token_pieces_from_env():
    return os.environ.get("STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN", None)


def get_github_token_workflows_from_env():
    return os.environ.get("STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS", None)


def get_workflows_repository_from_env():
    return os.environ.get("STARFLOW_GITHUB_WORKFLOWS_REPOSITORY", None)


def get_registry_token_from_env():
    return os.environ.get('GHCR_PASSWORD', "")


def get_github_token_pieces_from_config_or_env():
    if Path('config-starflow-local.toml').is_file():
        with open("config-starflow-local.toml", "rb") as f:
            config = tomli.load(f)
        return config.get('github').get('STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN', None)

    return os.environ.get("STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN", None)

@click.command()
@click.option(
    '--cluster-name',
    prompt='Local cluster name',
    default=get_cluster_name_from_env,
    help='Define the name for the local k8s cluster.'
)
@click.option(
    '--workflows-repository',
    prompt='Workflows repository',
    default=get_workflows_repository_from_env,
    help='Github repository where the starflow workflows will be stored.'
)
@click.option(
    '--github-workflows-ssh-private-key',
    prompt='Github ssh private for Workflows repository. If none, it will create a ssh key pair to be used.',
    default=get_github_workflows_ssh_private_key_from_env,
    help='Github ssh key for GitSync read/write operations on Workflows repository. The private key will be used in the starflow cluster and the public key should be added to the Github repository deploy keys.'
)
@click.option(
    '--github-default-pieces-repository-token',
    prompt='Github token for Pieces repository',
    default=get_github_token_pieces_from_env,
    help='Github token for read operations on Pieces repositories.'
)
@click.option(
    '--github-workflows-token',
    prompt='Github token for Workflows Repository',
    default=get_github_token_workflows_from_env,
    help='Github token for read/write operations on Workflows Repository.'
)
@click.option(
    '--deploy-mode',
    prompt='Deploy mode',
    default="local-k8s",
    help='Deploy mode - either "local" or "remote".'
)
@click.option(
    '--local-pieces-repository-path',
    prompt='Local pieces repository path. Example: ["/path/to/repo1", "/path/to/"repo2"]',
    default=[],
    help='List of local pieces repository paths.',
)
@click.option(
    "--local-starflow-path",
    prompt="Local starflow path",
    default="",
    help="Local starflow path. It is used only if dev-mode is set to 'local' and it will allow you to use hot reloading for local starflow.",
)
@click.option(
    '--local-rest-image',
    prompt='Local starflow REST image to use if local-k8s-dev deploy mode.',
    default="",
    help='Local starflow REST image to use if local-k8s-dev deploy mode.',
)
@click.option(
    '--local-frontend-image',
    prompt='Local starflow Front image to use if local-k8s-dev deploy mode.',
    default="",
    help='Local starflow Front image to use if local-k8s-dev deploy mode.',
)
@click.option(
    '--local-airflow-image',
    prompt='Local starflow Airflow image to use if local-k8s-dev deploy mode.',
    default="",
    help='Local starflow Airflow image to use if local-k8s-dev deploy mode.',
)
def cli_prepare_platform(
    cluster_name,
    workflows_repository,
    github_workflows_ssh_private_key,
    github_default_pieces_repository_token,
    github_workflows_token,
    deploy_mode,
    local_pieces_repository_path,
    local_starflow_path,
    local_rest_image,
    local_frontend_image,
    local_airflow_image

):
    """Prepare local folder for running a starflow platform."""
    platform.prepare_platform(
        cluster_name=cluster_name,
        workflows_repository=workflows_repository,
        github_workflows_ssh_private_key=github_workflows_ssh_private_key,
        github_default_pieces_repository_token=github_default_pieces_repository_token,
        github_workflows_token=github_workflows_token,
        deploy_mode=deploy_mode,
        local_pieces_repository_path=ast.literal_eval(local_pieces_repository_path),
        local_starflow_path=local_starflow_path,
        local_rest_image=local_rest_image,
        local_frontend_image=local_frontend_image,
        local_airflow_image=local_airflow_image
    )


@click.command()
@click.option(
    "--install-airflow",
    default=True,
    help="Install Airflow services."
)
@click.option(
    "--use-gpu",
    is_flag=True,
    # Nvidia operator plugin reference: https://catalog.ngc.nvidia.com/orgs/nvidia/containers/gpu-operator
    help="Allow the platform to use GPUs. It will install NVIDIA plugins.",
    default=False
)
def cli_create_platform(install_airflow, use_gpu):
    """Create cluster, install services and run starflow platform."""
    platform.create_platform(install_airflow, use_gpu)


@click.command()
def cli_destroy_platform():
    """Destroy Kind cluster."""
    platform.destroy_platform()


@click.command()
@click.option(
    '--use-config-file',
    is_flag=True,
    help="Use config file to run platform.",
    default=False
)
@click.option(
    '--dev',
    is_flag=True,
    help="Run platform in dev mode.",
    default=False
)
@click.option(
    '--debug',
    is_flag=True,
    help="Debug mode prints docker compose messages on terminal.",
    default=False
)
@click.option(
    '--stop',
    is_flag=True,
    help="Stop and remove containers.",
    default=False
)
@click.option(
    '--github-token',
    prompt='Github token for access default pieces repositories.',
    help='Github token for access default pieces repositories.',
    default=get_github_token_pieces_from_config_or_env,
)
def cli_run_platform_compose(use_config_file, dev, debug, stop, github_token):
    """Run starflow platform locally with docker compose. Do NOT use this in production."""
    if stop:
        platform.stop_platform_compose()
    else:
        platform.run_platform_compose(github_token=github_token, use_config_file=use_config_file, dev=dev, debug=debug)


@click.command()
def cli_stop_platform_compose():
    """Stop starflow platform locally with docker compose. Do NOT use this in production."""
    platform.stop_platform_compose()


@click.group()
@click.pass_context
def cli_platform(ctx):
    """starflow platform actions"""
    if ctx.invoked_subcommand == "prepare":
        console.print("Let's get you started configuring a local starflow platform:")
    elif ctx.invoked_subcommand == "create":
        console.print("Your local starflow platform is being created. This might take a while...")


cli_platform.add_command(cli_prepare_platform, name="prepare")
cli_platform.add_command(cli_create_platform, name="create")
cli_platform.add_command(cli_destroy_platform, name="destroy")
cli_platform.add_command(cli_run_platform_compose, name="run-compose")
cli_platform.add_command(cli_stop_platform_compose, name="stop-compose")


###############################################################################
# PIECES REPOSITORY
###############################################################################

def generate_random_repo_name():
    return f"new_repository_{str(uuid.uuid4())[0:8]}"


@click.command()
@click.option(
    '--name',
    prompt="Repository's name",
    default=generate_random_repo_name,
    help="Repository's name"
)
@click.option(
    '--container-registry',
    prompt="Github Container Registry name",
    default="",
    help="Github container registry name"
)
def cli_create_piece_repository(name, container_registry):
    """Create a basic Pieces repository with placeholder files."""
    pieces_repository.create_pieces_repository(repository_name=name, container_registry=container_registry)


@click.command()
@click.option(
    '--build-images',
    is_flag=True,
    prompt='Build Docker images?',
    expose_value=True,
    default=False,
    help='If True (default), builds Docker images.'
)
@click.option(
    '--source-url',
    prompt='Url of source repository',
    default="",
    help='The base url for this Pieces repository.'
)
@click.option(
    '--tag-overwrite',
    default="",
    help='Overwrite tag for release.'
)
@click.option(
    '--dev',
    is_flag=True,
    default=False,
    help="Build pieces images using development base piece image."
)
def cli_organize_pieces_repository(build_images: bool, source_url: str, tag_overwrite: str, dev: bool):
    """Organize Pieces repository."""
    pieces_repository.organize_pieces_repository(build_images, source_url, tag_overwrite, dev)


@click.command()
@click.option(
    '--registry-token',
    prompt='Github Container Registry token',
    default=get_registry_token_from_env,
    help='Your Github Container Registry token with access to where the image will be published.'
)
def cli_publish_images(registry_token: str):
    """Publish images to github container registry from mapping."""
    if registry_token:
        os.environ['GHCR_PASSWORD'] = registry_token
    console.print("Using registry token to publish images")
    pieces_repository.publish_docker_images()


@click.command()
@click.option(
    '--tag-name',
    default="",
    help='Tag name'
)
@click.option(
    '--commit-sha',
    default="",
    help='Commit SHA'
)
def cli_create_release(tag_name: str, commit_sha: str):
    """
    Get release version for the Pieces repository in github stdout format.
    Used by github actions to set the release version.
    Needs the following env vars:
        - GITHUB_TOKEN
        - GITHUB_REPOSITORY
    """
    pieces_repository.create_release(tag_name=tag_name, commit_sha=commit_sha)


@click.command()
@click.option(
    '--tag-name',
    help='Tag name'
)
def cli_delete_release(tag_name: str):
    """
    Used by github actions to delete a release with given tag name.
    Needs the following env vars:
        - GITHUB_TOKEN
        - GITHUB_REPOSITORY
    """
    pieces_repository.delete_release(tag_name=tag_name)


@click.group()
@click.pass_context
def cli_piece(ctx):
    """Pieces repository actions"""
    if ctx.invoked_subcommand == "organize":
        console.print(f"Organizing Pieces Repository at: {Path('.').resolve()}")
    elif ctx.invoked_subcommand == "create":
        pass


cli_piece.add_command(cli_organize_pieces_repository, name="organize")
cli_piece.add_command(cli_create_piece_repository, name="create")
cli_piece.add_command(cli_create_release, name="release")
cli_piece.add_command(cli_delete_release, name="delete-release")
cli_piece.add_command(cli_publish_images, name="publish-images")


###############################################################################
# PARENT GROUP
###############################################################################

@click.command()
def cli_run_piece_k8s():
    """Run Piece on Kubernetes Pod"""
    from starflow.scripts.run_piece_docker import run_piece as run_piece_in_docker
    console.print("Running Piece inside K8s pod...")
    run_piece_in_docker()

@click.command()
def cli_run_piece_branch_k8s():
    """Run Piece Branch on Kubernetes Pod"""
    from starflow.scripts.run_piece_branch_docker import run_piece_branch as run_piece_in_docker
    console.print("Running Piece inside K8s pod...")
    run_piece_in_docker()


@click.command()
def cli_run_piece_docker():
    """Run Piece on Docker container"""
    from starflow.scripts.run_piece_docker import run_piece as run_piece_in_docker
    console.print('Running Piece inside Docker container...')
    run_piece_in_docker()


###############################################################################
# PARENT GROUP
###############################################################################

@click.group()
@click.version_option(starflow.__version__)
@click.pass_context
def cli(ctx):
    # console.print(msg, style="rgb(109,125,176)", highlight=False)
    console.print("")
    console.print("Welcome to starflow! :red_heart-emoji:")
    console.print("")


cli.add_command(cli_platform, name="platform")
cli.add_command(cli_piece, name="piece")
cli.add_command(cli_run_piece_k8s, name="run-piece-k8s")
cli.add_command(cli_run_piece_branch_k8s, name="run-piece-branch-k8s")
cli.add_command(cli_run_piece_docker, name='run-piece-docker')


if __name__ == '__main__':
    cli()



from github import Github


class GithubRestClient(Github):
    def __init__(self, token: str):
        super().__init__(token)

    def get_releases(self, repo_name: str):
        """
        Get all releases from a repository.
        """
        repo = super().get_repo(repo_name)
        releases = [e for e in repo.get_releases()]
        return releases

    def get_tag(self, repo_name: str, tag_name: str):
        """
        Get a tag from a repository.
        """
        repo = super().get_repo(repo_name)
        tags = repo.get_tags()
        for tag in tags:
            if str(tag.name) == tag_name:
                return tag
        return None

    def list_contents(self, repo_name: str, folder_path: str):
        """
        List the contents of a folder from a repository.
        """
        repo = super().get_repo(repo_name)
        return repo.get_contents(folder_path)

    def get_contents(self, repo_name: str, file_path: str):
        """
        Get the contents of a file from a repository.
        """
        repo = super().get_repo(repo_name)
        contents = repo.get_contents(file_path)
        return contents

    def create_file(self, repo_name: str, file_path: str, content: str):
        """
        Create a file in a repository.
        """
        repo = super().get_repo(repo_name)
        commit_msg = 'Create file'
        repo.create_file(
            path=file_path,
            message=commit_msg,
            content=content
        )

    def get_commits(self, repo_name: str, number_of_commits: int = 1):
        """
        Get all commits from the repository.
        """
        repo = super().get_repo(repo_name)
        commits = [e for e in repo.get_commits()[:number_of_commits]]
        return commits

    def get_commit(self, repo_name: str, commit_sha: str):
        repo = super().get_repo(repo_name)
        commit = repo.get_commit(commit_sha)
        return commit

    def compare_commits(
        self,
        repo_name: str,
        base_sha: str,
        head_sha: str
    ):
        """
        Compare two commits and return the diff.
        """
        repo = super().get_repo(repo_name)
        diff = repo.compare(base_sha, head_sha)
        return diff

    def create_release(
        self,
        repo_name: str,
        version: str,
        tag_message: str,
        release_message: str,
        target_commit_sha: str,
        release_type: str = "commit"
    ):
        """
        Create a release and tag in a repository.
        """
        repo = super().get_repo(repo_name)
        release = repo.create_git_tag_and_release(
            tag=version,
            tag_message=tag_message,
            release_name=version,
            release_message=release_message,
            object=target_commit_sha,
            type=release_type
        )
        return release

    def delete_release_by_tag(self, repo_name: str, tag_name: str):
        """
        Delete a release with a specific tag in a repository.
        """
        repo = super().get_repo(repo_name)
        try:
            # Find the release by tag name
            release_to_delete = None
            for release in repo.get_releases():
                if release.tag_name == tag_name:
                    release_to_delete = release
                    break

            if release_to_delete:
                # Delete the release
                release_to_delete.delete_release()
        except Exception as e:
            raise Exception(f"An error occurred: {e}")

    def delete_tag(self, repo_name: str, tag_name: str):
        """
        Delete a tag from a GitHub repository.
        """
        repo = super().get_repo(repo_name)
        try:
            # Delete tag by reference
            ref = 'tags/' + tag_name
            repo.get_git_ref(ref).delete()
            print(f"Tag '{tag_name}' deleted successfully.")
        except Exception as e:
            raise Exception(f"Error deleting tag: {e}")



from typing import Dict, Optional
from airflow.utils.context import Context

from starflow.client.starflow_backend_client import starflowBackendRestClient
from starflow.schemas.shared_storage import WorkflowSharedStorage


class BasestarflowOperator:
    """
    This class implements common operations for all starflow Operators running under a Task.
    DEPRECATED - delete this later 
    """

    def __init__(
        self, 
        task_id: str, 
        piece_name: str,  
        deploy_mode: str,
        repository_id: int,
        piece_input_kwargs: Optional[Dict] = None, 
        workflow_shared_storage: WorkflowSharedStorage = None,
        starflow_client_url: Optional[str] = None,
    ):
        self.task_id = task_id
        self.piece_name = piece_name
        self.deploy_mode = deploy_mode
        self.repository_id = repository_id
        self.piece_input_kwargs = piece_input_kwargs
        self.workflow_shared_storage = workflow_shared_storage
        if starflow_client_url is None:
            starflow_client_url = "http://starflow-rest:8000/"
        self.starflow_client = starflowBackendRestClient(base_url=starflow_client_url)

    def _get_piece_secrets(self, piece_repository_id: int, piece_name: str):
        """Get piece secrets values from starflow API"""
        secrets_response = self.starflow_client.get_piece_secrets(
            piece_repository_id=piece_repository_id,
            piece_name=piece_name
        )
        if secrets_response.status_code != 200:
            raise Exception(f"Error getting piece secrets: {secrets_response.json()}")
        piece_secrets = {
            e.get('name'): e.get('value') 
            for e in secrets_response.json()
        }
        return piece_secrets
    
    @staticmethod
    def _get_upstream_xcom_data_from_task_ids(task_ids: list, context: Context):
        upstream_xcoms_data = dict()
        for tid in task_ids:
            upstream_xcoms_data[tid] = context['ti'].xcom_pull(task_ids=tid)
        return upstream_xcoms_data


from .container_resources import ContainerResourcesModel
from .deploy_mode import DeployModeType
from .display_result import DisplayResultFileType
from .piece_metadata import PieceMetadata
from .shared_storage import (
    WorkflowSharedStorage, 
    StorageSource,
    shared_storage_map
)



from pathlib import Path
import os
import json
from typing import Any, Dict, List, Optional
from airflow.operators.python import PythonOperator as AirflowPythonOperator
from airflow.utils.context import Context

from starflow.scripts.load_piece import load_piece_class_from_path, load_piece_models_from_path

# DEPRECATED
def make_python_callable(operator_name, deploy_mode, task_id, dag_id):
    # Import Operator, already configured with metadata, from mounted volume
    volume_mount_path = os.getenv("VOLUME_MOUNT_PATH_DOCKER", "/opt/mnt/fs")
    operators_folder_path = Path(volume_mount_path) / f"code_repository/operators"
    compiled_metadata_path = volume_mount_path / "code_repository/.flowui/compiled_metadata.json"
    with open(str(compiled_metadata_path), "r") as f:
        compiled_metadata = json.load(f)

    operator_class = load_piece_class_from_path(
        operators_folder_path=operators_folder_path.resolve(),
        operator_name=operator_name,
        operator_metadata=compiled_metadata[operator_name]
    )
    operator_object = operator_class(
        deploy_mode=deploy_mode,
        task_id=task_id,
        dag_id=dag_id,
    )

    input_model_class, output_model_class, secrets_model_class = load_piece_models_from_path(
        operators_folder_path=operators_folder_path,
        operator_name=operator_name
    )
    return operator_object.run_piece_function, input_model_class, output_model_class, secrets_model_class


class PythonOperator(AirflowPythonOperator):

    def __init__(
        self, 
        *, 
        op_args: Optional[List] = None, 
        op_kwargs: Optional[Dict] = None, 
        templates_dict: Optional[Dict] = None, 
        templates_exts: Optional[List[str]] = None, 
        show_return_value_in_logs: bool = True,
        make_python_callable_kwargs: Dict,  # Used for setting up the custom operators at execute_callable()
        **kwargs
    ) -> None:
        self.make_python_callable_kwargs = make_python_callable_kwargs
        super().__init__(
            # *,
            python_callable=make_python_callable, 
            op_args=op_args, 
            op_kwargs=op_kwargs, 
            templates_dict=templates_dict, 
            templates_exts=templates_exts, 
            show_return_value_in_logs=show_return_value_in_logs,
            **kwargs
        )
    
    def execute(self, context: Context) -> Any:
        # THIS WAS CHANGED 06/07/2022 to accomodate starflow format for calling run_piece_function

        # context_merge(context, self.op_kwargs, templates_dict=self.templates_dict)
        # self.op_kwargs = self.determine_kwargs(context)

        self.context = context
        self.piece_function, self.operator_input_model_class, self.operator_output_model_class, self.operator_secrets_model_class = self.python_callable(**self.make_python_callable_kwargs)

        return_value = self.execute_callable()

        if self.show_return_value_in_logs:
            self.log.info("Done. Returned value was: %s", return_value)
        else:
            self.log.info("Done. Returned value not shown")

        return return_value

    def execute_callable(self):
        return self.piece_function(
            operator_input_model=self.operator_input_model_class, 
            operator_output_model=self.operator_output_model_class, 
            operator_secrets_model=self.operator_secrets_model_class,
            op_kwargs=self.op_kwargs, 
            airflow_context=self.context
        )


from pathlib import Path
import importlib
import json
import sys
import os
from typing import Optional
from starflow.logger import get_configured_logger


def piece_dry_run(
    piece_name: str,
    input_data: dict,
    repository_folder_path: Optional[str] = None,
    secrets_data: dict = None,
    results_path: str = None
) -> dict:
    """
    Example usage of dry run:

    from starflow.testing.dry_run import piece_dry_run

    piece_dry_run(
        repository_folder_path=".",
        piece_name="SingleLogPiece",
        input_data={
            "message": "testing msg",
            "input_str": 'testing',
            "input_int": 1,
            ...
        }
    )
    """
    logger = get_configured_logger("piece_dry_run")

    pieces_images_map = os.environ.get("PIECES_IMAGES_MAP", {})
    http_server = None
    if pieces_images_map and piece_name in pieces_images_map:
        try:
            from starflow.testing.http_client import TestingHttpClient

            logger.info('Running pieces dry run with http client')
            http_client = TestingHttpClient()
            pieces_images_map = json.loads(pieces_images_map)
            piece_image = pieces_images_map.get(piece_name, None)
            if not piece_image:
                raise Exception(f"Piece {piece_name} not found in PIECES_IMAGES_MAP")
            http_server = http_client.start_http_server(image=piece_image)
            dry_run_response = http_client.send_dry_run_request(piece_name, input_data, secrets_data)
            http_server.stop()
            http_server.remove()
            return dry_run_response.json()
        except Exception as e:
            logger.error(f"Error running dry run with http client: {e}")
            if http_server:
                http_server.stop()
                http_server.remove()
            raise e

    if not repository_folder_path:
        repository_folder_path = '.'
    if not results_path:
        results_path = Path('./dry_run_results')
        results_path.mkdir(parents=True, exist_ok=True)

    pieces_folder_path = str(Path(repository_folder_path).resolve() / "pieces")
    if pieces_folder_path not in sys.path:
        sys.path.append(pieces_folder_path)

    # Load Operator class
    importlib.invalidate_caches()
    piece_module = importlib.import_module(f"{piece_name}.piece")
    piece_class = getattr(piece_module, piece_name)

    # Load metadata
    metadata_path = Path(pieces_folder_path) / f"{piece_name}/metadata.json"
    with open(metadata_path, "r") as f:
        metadata = json.load(f)

    # Set Operator class metadata
    piece_class.set_metadata(metadata=metadata)

    # Load Operator Models
    importlib.invalidate_caches()
    piece_model_module = importlib.import_module(f"{piece_name}.models")
    piece_input_model_class = getattr(piece_model_module, "InputModel")
    piece_secrets_model_class = getattr(piece_model_module, "SecretsModel", None)
    piece_output_model_class = getattr(piece_model_module, "OutputModel", None)

    # Dry run Operator
    output = piece_class.dry_run(
        input_data=input_data,
        piece_input_model=piece_input_model_class,
        piece_output_model=piece_output_model_class,
        piece_secrets_model=piece_secrets_model_class,
        secrets_data=secrets_data,
        results_path=str(results_path)
    )
    return json.loads(output.model_dump_json())



from airflow.providers.docker.operators.docker import DockerOperator, Mount
from airflow.utils.context import Context
from typing import Dict, Optional, Any
import os

from starflow.client.starflow_backend_client import starflowBackendRestClient
from starflow.schemas import WorkflowSharedStorage, StorageSource
from docker.types import Mount
import docker
class starflowDockerOperator(DockerOperator):

    def __init__(
        self,
        dag_id: str,
        task_id: str,
        piece_name: str,
        deploy_mode: str,  # TODO enum
        repository_url: str,
        repository_version: str,
        workspace_id: int,
        piece_input_kwargs: Optional[Dict] = None,
        workflow_shared_storage: WorkflowSharedStorage = None,
        container_resources: Optional[Dict] = None,
        **docker_operator_kwargs
    ) -> None:
        self.task_id = task_id
        self.piece_name = piece_name
        self.deploy_mode = deploy_mode
        self.repository_url = repository_url
        self.repository_version = repository_version
        self.workspace_id = workspace_id
        self.piece_input_kwargs = piece_input_kwargs
        self.workflow_shared_storage = workflow_shared_storage
        self.container_resources = container_resources or {}

        # Environment variables
        self.environment = {
            "STARFLOW_PIECE": piece_name,
            "STARFLOW_INSTANTIATE_PIECE_KWARGS": str({
                "deploy_mode": deploy_mode,
                "task_id": task_id,
                "dag_id": dag_id,
            }),
            "STARFLOW_RUN_PIECE_KWARGS": str(piece_input_kwargs),
            "STARFLOW_WORKFLOW_SHARED_STORAGE": self.workflow_shared_storage.model_dump_json() if self.workflow_shared_storage else "",
            "AIRFLOW_CONTEXT_EXECUTION_DATETIME": "{{ dag_run.logical_date | ts_nodash }}",
            "AIRFLOW_CONTEXT_DAG_RUN_ID": "{{ run_id }}",
        }

        # Shared Storage variables
        self.shared_storage_base_mount_path = '/home/shared_storage'
        self.shared_storage_upstream_ids_list = list()
        shared_storage_host_path = os.environ.get('LOCAL_STARFLOW_SHARED_DATA_PATH', '')
        shared_storage_container_path = '/home/shared_storage'
        mounts = []

        # TODO remove - used in DEV only #######################
        dev_pieces = False
        if dev_pieces:
            piece_repo_name = repository_url.split("/")[-1]
            #local_repos_path = f"/mnt/shared_storage/Github/{piece_repo_name}"
            local_repos_path = f"/home/vinicius/Documents/work/tauffer/{piece_repo_name}"
            mounts = [
                # TODO remove
                Mount(source='/home/vinicius/Documents/work/tauffer/starflow/src/starflow', target='/usr/local/lib/python3.10/site-packages/starflow/', type='bind', read_only=True),
                #Mount(source='/mnt/shared_storage/Github/starflow/src/starflow', target='/usr/local/lib/python3.10/site-packages/starflow/', type='bind', read_only=True),
                Mount(source=local_repos_path, target='/home/starflow/pieces_repository/', type='bind', read_only=True),
            ]
        ########################################################

        if self.workflow_shared_storage and str(self.workflow_shared_storage.source.value).lower() == str(getattr(StorageSource, 'local').value).lower():
            mounts.append(
                Mount(
                    source=shared_storage_host_path,
                    target=shared_storage_container_path,
                    type='bind',
                    read_only=False
                ),
            )

        self.device_requests = []
        if self.container_resources.get('use_gpu', False):
            self.device_requests=[
                docker.types.DeviceRequest(count=-1, capabilities=[['gpu']])
            ]
        super().__init__(
            **docker_operator_kwargs,
            task_id=task_id,
            docker_url='tcp://docker-proxy:2375',
            mounts=mounts,
            device_requests=self.device_requests,
            environment=self.environment,
        )

    def _get_piece_secrets(self) -> Dict[str, Any]:
        """Get piece secrets values from starflow API"""
        params = {
            "workspace_id": self.workspace_id,
            "url": self.repository_url,
            "version": self.repository_version,
            "page": 0,
            "page_size": 1,
        }

        piece_repository_data = self.starflow_client.get_piece_repositories_from_workspace_id(
            params=params
        ).json()
        secrets_response = self.starflow_client.get_piece_secrets(
            piece_repository_id=piece_repository_data["data"][0]["id"],
            piece_name=self.piece_name
        )
        if secrets_response.status_code != 200:
            raise Exception(f"Error getting piece secrets: {secrets_response.json()}")
        piece_secrets = {}
        for e in secrets_response.json():
            if not e.get('value') and not e.get('required'):
                continue
            piece_secrets[e.get('name')] = e.get('value')

        return piece_secrets

    @staticmethod
    def _get_upstream_xcom_data_from_task_ids(task_ids: list, context: Context):
        upstream_xcoms_data = dict()
        for tid in task_ids:
            upstream_xcoms_data[tid] = context['ti'].xcom_pull(task_ids=tid)
        return upstream_xcoms_data

    def _get_piece_kwargs_value_from_upstream_xcom(
        self,
        value: Any
    ):
        if isinstance(value, dict) and value.get("type") == "fromUpstream":
            upstream_task_id = value["upstream_task_id"]
            output_arg = value["output_arg"]
            if upstream_task_id not in self.shared_storage_upstream_ids_list:
                self.shared_storage_upstream_ids_list.append(upstream_task_id)
            return self.upstream_xcoms_data[upstream_task_id][output_arg]
        elif isinstance(value, list):
            return [self._get_piece_kwargs_value_from_upstream_xcom(item) for item in value]
        elif isinstance(value, dict):
            return {
                k: self._get_piece_kwargs_value_from_upstream_xcom(v)
                for k, v in value.items()
            }
        return value

    def _update_piece_kwargs_with_upstream_xcom(self):
        if not self.piece_input_kwargs:
            self.piece_input_kwargs = dict()

        updated_piece_kwargs = {
            k: self._get_piece_kwargs_value_from_upstream_xcom(
                value=v
            ) for k, v in self.piece_input_kwargs.items()
        }
        self.piece_input_kwargs = updated_piece_kwargs
        self.environment['AIRFLOW_UPSTREAM_TASKS_IDS_SHARED_STORAGE'] = str(self.shared_storage_upstream_ids_list)
        self.environment['STARFLOW_RUN_PIECE_KWARGS'] = str(self.piece_input_kwargs)

    def _prepare_execute_environment(self, context: Context):
        """
        Prepare execution with the following configurations:
        - pass extra arguments and configuration as environment variables to the pod
        - add shared storage sidecar container to the pod - if shared storage is FUSE based
        - add shared storage volume mounts to the pod - if shared storage is NFS based or local
        """
        # Fetch upstream tasks ids and save them in an ENV var
        upstream_task_ids = [t.task_id for t in self.get_direct_relatives(upstream=True)]
        self.environment['AIRFLOW_UPSTREAM_TASKS_IDS'] = str(upstream_task_ids)
        self.environment['STARFLOW_WORKFLOW_SHARED_STORAGE_SOURCE_NAME'] = str(self.workflow_shared_storage.source.name) if self.workflow_shared_storage else None

        # Save updated piece input kwargs with upstream data to environment variable
        self.upstream_xcoms_data = self._get_upstream_xcom_data_from_task_ids(task_ids=upstream_task_ids, context=context)
        self._update_piece_kwargs_with_upstream_xcom()

        piece_secrets = self._get_piece_secrets()
        self.environment['STARFLOW_PIECE_SECRETS'] = str(piece_secrets)

        dag_id = context["dag_run"].dag_id
        dag_run_id = context['run_id']
        dag_run_id_path = dag_run_id.replace("-", "_").replace(".", "_").replace(" ", "_").replace(":", "_").replace("+", "_")
        self.workflow_run_subpath = f"{dag_id}/{dag_run_id_path}"
        self.environment['STARFLOW_WORKFLOW_RUN_SUBPATH'] = self.workflow_run_subpath

    def execute(self, context: Context) -> Optional[str]:
        """
        Code from here onward is executed by the Worker and not by the Scheduler.
        """
        self.starflow_client = starflowBackendRestClient(base_url="http://starflow-rest:8000/")
        # env var format = {"name": "value"}
        self._prepare_execute_environment(context=context)
        result = super().execute(context=context)
        self._shared_storage_usage_in_bytes = result.get('_shared_storage_usage_in_bytes', 0)
        return result



from pathlib import Path
import importlib
import sys


def load_piece_class_from_path(pieces_folder_path: str, piece_name: str, piece_metadata: dict = None):
    # Import Operator from path, then set up piece module and class
    pieces_folder_path = str(Path(pieces_folder_path).resolve())
    if pieces_folder_path not in sys.path:
        sys.path.append(pieces_folder_path)

    importlib.invalidate_caches()
    piece_module = importlib.import_module(f"{piece_name}.piece")
    piece_class = getattr(piece_module, piece_name)

    # Set Operator class metadata
    if piece_metadata:
        piece_class.set_metadata(metadata=piece_metadata)

    return piece_class


def load_piece_models_from_path(pieces_folder_path: str, piece_name: str):
    # Import Operator from mounted volume, then set up piece module and class
    pieces_folder_path = str(Path(pieces_folder_path).resolve())
    if pieces_folder_path not in sys.path:
        sys.path.append(pieces_folder_path)

    importlib.invalidate_caches()
    piece_model_module = importlib.import_module(f"{piece_name}.models")
    piece_input_model_class = getattr(piece_model_module, "InputModel")
    piece_output_model_class = getattr(piece_model_module, "OutputModel")
    piece_secrets_model_class = getattr(piece_model_module, "SecretsModel", None)

    return piece_input_model_class, piece_output_model_class, piece_secrets_model_class

def load_piece_branch_models_from_path(pieces_folder_path: str, piece_name: str):
    # Import Operator from mounted volume, then set up piece module and class
    pieces_folder_path = str(Path(pieces_folder_path).resolve())
    if pieces_folder_path not in sys.path:
        sys.path.append(pieces_folder_path)

    importlib.invalidate_caches()
    piece_model_module = importlib.import_module(f"{piece_name}.models")
    piece_input_model_class = getattr(piece_model_module, "InputModel")

    return piece_input_model_class


from pydantic import BaseModel, Field
from enum import Enum
from typing import Optional


class StorageSource(str, Enum):
    none = "None"
    local = 'Local'
    aws_s3 = "AWS S3"


class StorageMode(str, Enum):
    none = 'None'
    read = 'Read'
    write = 'Write'
    read_write = 'Read/Write'


class WorkflowSharedStorage(BaseModel):
    source: StorageSource = Field(
        description="The service to be used as the source of storage.", 
        default="none",
    )
    mode: StorageMode = Field(
        description="The access mode to the storage source.", 
        default="none"
    )
    storage_piece_name: str = Field(
        description="The name of the default piece to be used for the storage source.",
    )



class LocalSharedStorage(WorkflowSharedStorage):
    source: StorageSource = StorageSource.local
    storage_piece_name: str = "LocalStoragePiece" # TODO to be implemented


class AwsS3SharedStorage(WorkflowSharedStorage):
    source: StorageSource = StorageSource.aws_s3
    bucket: str = Field(
        description="The name of the bucket to be used as the root of the storage source."
    )
    storage_piece_name: str = "AWSS3StoragePiece"
    base_folder: str = Field(
        description="The base folder to be used as the root of the storage source.",
        default="",
    )



shared_storage_map = {
    "none": None,
    "local": LocalSharedStorage,
    "aws_s3": AwsS3SharedStorage,
}


from airflow.operators.python import ExternalPythonOperator as AirflowExternalPythonOperator
from airflow.utils.context import Context
from pathlib import Path
import os
from starflow.client.starflow_backend_client import starflowBackendRestClient
import json
from typing import Any, Dict, List, Optional
from starflow.scripts.load_piece import load_piece_class_from_path, load_piece_models_from_path


def make_python_callable(piece_name, task_id, deploy_mode, dag_id):
    pieces_repository_path = Path('/opt/airflow/starflow/pieces_repository')
    pieces_folder = pieces_repository_path / f"pieces"
    compiled_metadata_path = pieces_repository_path / ".starflow/compiled_metadata.json"

    with open(str(compiled_metadata_path), "r") as f:
        compiled_metadata = json.load(f)

    piece_class = load_piece_class_from_path(
        pieces_folder_path=pieces_folder.resolve(),
        piece_name=piece_name,
        piece_metadata=compiled_metadata[piece_name]
    )
    piece_object = piece_class(
        deploy_mode=deploy_mode,
        task_id=task_id,
        dag_id=dag_id,
    )

    input_model_class, output_model_class, secrets_model_class = load_piece_models_from_path(
        pieces_folder_path=pieces_folder,
        piece_name=piece_name
    )
    return piece_object, input_model_class, output_model_class, secrets_model_class

class starflowExternalPythonOperator(AirflowExternalPythonOperator):
    def __init__(
        self,
        piece_name: str,
        repository_id: int,
        piece_kwargs: Optional[Dict] = None, 
        make_python_callable_kwargs: Dict = None,  # Used for setting up the custom pieces at execute_callable()
        venv_path=None,
        **kwargs
    ) -> None:
        self.venv_path = venv_path
        self.piece_kwargs = piece_kwargs
        self.make_python_callable_kwargs = make_python_callable_kwargs
        self.backend_client = starflowBackendRestClient(base_url="http://starflow-rest:8000/")
        self.running_piece_name = piece_name
        self.repository_id = repository_id
        if self.venv_path is None:
            self.venv_path = '/opt/airflow/starflow/venv/bin/python'
        super().__init__(
            python=self.venv_path,
            python_callable=make_python_callable,
            **kwargs
        )


    def _get_piece_secrets(self, piece_repository_id: int, piece_name: str):
        # Get piece secrets values from api and append to env vars
        secrets_response = self.backend_client.get_piece_secrets(
            piece_repository_id=piece_repository_id,
            piece_name=piece_name
        )
        if secrets_response.status_code != 200:
            raise Exception(f"Error getting piece secrets: {secrets_response.json()}")
        piece_secrets = {
            e.get('name'): e.get('value') 
            for e in secrets_response.json()
        }
        return piece_secrets

    def _update_piece_kwargs_with_upstream_xcom(self, upstream_xcoms_data: dict):
        #starflow_run_op_kwargs = [var for var in self.env_vars if getattr(var, 'name', None) == 'STARFLOW_RUN_PIECE_KWARGS']
        #starflow_run_piece_kwargs = self.piece_kwargs
        if not self.piece_kwargs:
            self.piece_kwargs = dict()

        # Update Operator kwargs with upstream tasks XCOM data
        # Also updates the list of upstream tasks for which we need to mount the results path
        updated_op_kwargs = dict()
        for k, v in self.piece_kwargs.items():
            if isinstance(v, dict) and v.get("type", None) == "fromUpstream":
                upstream_task_id = v.get("upstream_task_id")
                output_arg = v.get("output_arg")
                output_value = upstream_xcoms_data[upstream_task_id][output_arg]
                output_type = upstream_xcoms_data[upstream_task_id][f"{output_arg}_type"]
                # If upstream output type is FilePath or DirectoryPath, we need to add the basic path prefix
                # if output_type in ["file-path", "directory-path"]:
                #     output_value = f"{self.shared_storage_base_mount_path}/{upstream_task_id}/results/{output_value}"
                updated_op_kwargs[k] = output_value
                if upstream_task_id not in self.shared_storage_upstream_ids_list:
                    self.shared_storage_upstream_ids_list.append(upstream_task_id)
            else:
                updated_op_kwargs[k] = v
        self.piece_kwargs = updated_op_kwargs

    def _get_upstream_xcom_data(self, context: Context):
        upstream_task_ids = [t.task_id for t in self.get_direct_relatives(upstream=True)]

        upstream_xcoms_data = dict()
        for tid in upstream_task_ids:
            upstream_xcoms_data[tid] = context['ti'].xcom_pull(task_ids=tid)
        return upstream_xcoms_data


    def execute(self, context: Context) -> Any:
        self.context = context
        upstream_xcom_data = self._get_upstream_xcom_data(context=context)

        self._update_piece_kwargs_with_upstream_xcom(upstream_xcoms_data=upstream_xcom_data)
        self.piece_object, self.piece_input_model_class, self.piece_output_model_class, self.piece_secrets_model_class = self.python_callable(**self.make_python_callable_kwargs)
        return_value = self.execute_callable()
        return return_value
    
    def execute_callable(self):
        piece_secrets = self._get_piece_secrets(piece_repository_id=self.repository_id, piece_name=self.running_piece_name)
        self.piece_object.run_piece_function(
            airflow_context=self.context,
            op_kwargs=self.piece_kwargs,
            piece_input_model=self.piece_input_model_class,
            piece_output_model=self.piece_output_model_class, 
            piece_secrets_model=self.piece_secrets_model_class,
            secrets_values=piece_secrets
        )
        
        return None


from pathlib import Path


def _get_version():
    version_file_path = Path(__file__).parent / 'VERSION'

    with version_file_path.open('r') as version_file:
        return version_file.read().strip() 

__version__ = _get_version()


from typing import Dict, Optional
from airflow.models import BaseOperator

from starflow.client.starflow_backend_client import starflowBackendRestClient


# This is  WIP, not working yet.

class starflowWorkerOperator(BaseOperator):
    """
    This Operator runs Pieces directly in a Worker.
    """

    def __init__(
        self, 
        dag_id: str,
        task_id: str, 
        piece_name: str, 
        deploy_mode: str, # TODO enum
        repository_id: int,
        piece_input_kwargs: Optional[Dict] = None, 
    ):
        self._get_piece_class()
        self._get_piece_secrets()
        self._get_airflow_conn_id()

    def _get_piece_class(self):
        # TODO
        pass

    def _get_airflow_conn_id(self):
        """
        Form correct conn_id string with conn_type + repository_id.
        Check if conn_id already exists in Airflow and, if not, create it.
        """
        # TODO
        conn_type = "aws"  # TODO: get from piece
        self.conn_id = f"{conn_type}_{self.repository_id}"
        try:
            response = self.backend_client.check_create_airflow_connection(
                conn_id=self.conn_id,
                conn_type=conn_type,
            )
        except Exception as e:
            raise e

    def execute(self, context):
        """Execute the Piece."""
        # TODO
        pass


import boto3
from botocore.exceptions import NoCredentialsError, UnknownCredentialError, NoAuthTokenError , ClientError
from starflow.logger import get_configured_logger

class S3StorageRepository:
    # Class to handle S3 starflow Storage support
    logger = get_configured_logger('S3StorageRepository')

    @classmethod
    def validate_s3_credentials_access(cls, access_key: str, secret_key: str, bucket: str) -> bool:
        cls.logger.info(f"Validating S3 Credentials for bucket {bucket}")
        try:
            s3 = boto3.client(
                's3',
                aws_access_key_id=access_key, 
                aws_secret_access_key=secret_key,
            )
            s3.list_objects(Bucket=bucket, MaxKeys=1)
            cls.logger.info(f"S3 Credentials are valid")
            return True
        except (NoCredentialsError, UnknownCredentialError, NoAuthTokenError, ClientError):
            cls.logger.info("Invalid credentials")
            return False
        except Exception as e:
            cls.logger.info(f"Error Validating credentials: {e}")
            return False
        return False



from pathlib import Path
import docker
import tomli
import json
import shutil
import os
from colorama import Fore, Style
from rich.console import Console


console = Console()


def publish_image(source_image_name: str):
    client = docker.from_env()
    print(f"Publishing docker image: {source_image_name}")
    print(Style.RESET_ALL + Style.DIM, end='')
    try:
        registry_url = 'https://ghcr.io'
        ghcr_username = os.environ.get("GHCR_USERNAME", None)
        ghcr_password = os.environ.get("GHCR_PASSWORD", None)
        client.login(username=ghcr_username, password=ghcr_password, registry=registry_url)
    except docker.errors.APIError:
        console.print("Unauthorized login")
        raise
    response = client.images.push(repository=source_image_name)
    print(response, end='')
    print(Style.RESET_ALL + Fore.BLUE + f"Finished publishing: {source_image_name}")


def build_image_from_tmp_dockerfile(
    source_image_name: str,
    path: str = ".",
    dockerfile: str = "Dockerfile-tmp"
):
    client = docker.from_env()
    try:
        os.environ["DOCKER_BUILDKIT"] = "1"
        print(Fore.BLUE + f"Building docker image: {source_image_name}")
        build_results = client.images.build(
            path=path,
            dockerfile=dockerfile,
            tag=source_image_name,
            nocache=True,
            forcerm=True
        )
        print(Style.RESET_ALL + Style.DIM, end='')
        for r in build_results[1]:
            if "stream" in r and r["stream"] != "\n":
                print(r)
        print(Style.RESET_ALL + Fore.BLUE + f"Finished building: {source_image_name}")
        print(Style.RESET_ALL)
    except Exception as e:
        raise Exception(e)
    finally:
        (Path(path) / dockerfile).unlink()


def build_images_from_pieces_repository(tag_overwrite: str | None = None, dev: bool = False):
    """
    Each dependencies group will need to have its own Docker image built and published to be used by starflow.
    This is because the Pieces source code goes baked in the images.
    """
    dependencies_path = Path(".") / "dependencies"
    starflow_path = Path(".") / ".starflow"
    config_path = Path(".") / "config.toml"

    # Get information from config.toml file
    with open(config_path, "rb") as f:
        repo_config = tomli.load(f).get("repository", {})

    # Tag overwrite
    if tag_overwrite:
        repo_config["VERSION"] = tag_overwrite

    docker_image_repository = repo_config.get("REPOSITORY_NAME")
    docker_image_version = repo_config.get("VERSION")
    github_container_registry_name = f'ghcr.io/{repo_config.get("REGISTRY_NAME")}'.lower()

    # Load dependencies_map.json file
    with open(starflow_path / "dependencies_map.json", "r") as f:
        pieces_dependencies_map = json.load(f)

    pieces_images_map = {}
    # Build docker images from unique definitions
    for group, v in pieces_dependencies_map.items():
        dependency_dockerfile = v["dependency"].get("dockerfile", None)
        dependency_requirements = v["dependency"].get("requirements_file", None)
        source_image_name = f"{github_container_registry_name}/{docker_image_repository}:{docker_image_version}-{group}"

        # If no extra dependency, use base Pod image and just copy the Pieces source code
        base_image = "ghcr.io/prochain-star-atlas/starflow-base-piece:latest"
        if dev:
            base_image = "ghcr.io/prochain-star-atlas/starflow-base-piece:latest-dev"

        if not any([dependency_dockerfile, dependency_requirements]):
            pieces_dependencies_map[group]["source_image"] = source_image_name
            dockerfile_str = f"""FROM {base_image}
COPY config.toml starflow/pieces_repository/
COPY pieces starflow/pieces_repository/pieces
COPY .starflow starflow/pieces_repository/.starflow
"""
            with open("Dockerfile-tmp", "w") as f:
                f.write(dockerfile_str)

        # If dependency is defined as a Dockerfile, copy it to root path and run build/publish
        elif dependency_dockerfile:
            pieces_dependencies_map[group]["source_image"] = source_image_name
            build_dockerfile_path = str(dependencies_path.resolve() / dependency_dockerfile)
            shutil.copyfile(build_dockerfile_path, "Dockerfile-tmp")

        # If dependency is defined as a requirements.txt
        elif dependency_requirements:
            pieces_dependencies_map[group]["source_image"] = source_image_name
            dockerfile_str = f"""FROM {base_image}
COPY config.toml starflow/pieces_repository/
COPY pieces starflow/pieces_repository/pieces
COPY .starflow starflow/pieces_repository/.starflow
COPY dependencies/{dependency_requirements} starflow/pieces_repository/dependencies/
RUN pip install --no-cache-dir -r starflow/pieces_repository/dependencies/{dependency_requirements}
"""
            with open("Dockerfile-tmp", "w") as f:
                f.write(dockerfile_str)

        for piece_name in v.get('pieces'):
            pieces_images_map[piece_name] = source_image_name

        build_image_from_tmp_dockerfile(source_image_name=source_image_name)

    var_value = json.dumps(pieces_images_map)
    os.environ["PIECES_IMAGES_MAP"] = var_value
    env_file = os.getenv('GITHUB_ENV')
    if env_file:
        with open(env_file, "a") as f:
            f.write(f"PIECES_IMAGES_MAP={var_value}")
    return pieces_dependencies_map



from pydantic import BaseModel, Field
from typing import List, Dict


class Dependency(BaseModel):
    docker_image: str = Field(
        description="Docker image to be used to run this Operator", 
        default=None
    )
    dockerfile: str = Field(
        description="Dockerfile to build the image for the container that will run this Operator",
        default=None
    )
    requirements_file: str = Field(
        description="Requirements file with pip packages to be installed in the container that will run this Operator",
        default=None
    )
    # TODO - validation: one and only one of these options


class InputSchema(BaseModel):
    title: str
    description: str
    type: str
    properties: Dict


class OutputSchema(BaseModel):
    title: str
    description: str
    type: str
    properties: Dict


class SecretsSchema(BaseModel):
    title: str
    description: str
    type: str
    properties: Dict


class PieceMetadata(BaseModel):
    name: str = Field(
        description="Piece name"
    )
    description: str = Field(
        description="Description of this Piece"
    )
    dependency: Dependency = None
    container_resources: Dict = None # TODO - add model for the container_resources dictionary
    tags: List[str] = None
    style: Dict = None # TODO - add model for the style dictionary
    input_schema: InputSchema = None
    output_schema: OutputSchema = None
    secrets_schema: SecretsSchema = None
    source_url: str = None


from starflow.base_piece import BaseBranchOutputModel
from starflow.scripts.load_piece import load_piece_branch_models_from_path, load_piece_class_from_path, load_piece_models_from_path
from pathlib import Path
import os
import ast
import json


def run_piece_branch():
     # Import Operator from File System, already configured with metadata
    piece_name = os.getenv("STARFLOW_PIECE")
    workspace_id = os.getenv("WORKSPACE_ID")

    pieces_repository_copy_path = Path("starflow/pieces_repository").resolve()
    pieces_folder_path = pieces_repository_copy_path / "pieces"
    compiled_metadata_path = pieces_repository_copy_path / ".starflow/compiled_metadata.json"

    with open(str(compiled_metadata_path), "r") as f:
        compiled_metadata = json.load(f)

    piece_class = load_piece_class_from_path(
        pieces_folder_path=pieces_folder_path,
        piece_name=piece_name,
        piece_metadata=compiled_metadata[piece_name]
    )

    piece_input_model_class = load_piece_branch_models_from_path(
        pieces_folder_path=pieces_folder_path,
        piece_name=piece_name
    )

    # Instantiate Operator
    instantiate_op_dict = ast.literal_eval(os.getenv("STARFLOW_INSTANTIATE_PIECE_KWARGS"))
    piece_object = piece_class(**instantiate_op_dict)

    # Run Operator
    run_piece_input_kwargs = ast.literal_eval(os.getenv("STARFLOW_RUN_PIECE_KWARGS"))
    piece_object.run_piece_function(
        piece_input_data=run_piece_input_kwargs,        
        piece_input_model=piece_input_model_class, 
        piece_output_model=BaseBranchOutputModel, 
        piece_secrets_model=None,
        workspace_id=workspace_id
    )

    return None


import re
import sys
import os
from packaging import version
import starflow
from starflow.client.github_rest_client import GithubRestClient


class Actions(object):
    github_client = GithubRestClient(token=os.environ.get("GITHUB_TOKEN"))
    repo_name = os.environ.get("GITHUB_REPOSITORY")
    new_tag = f"starflow-py-{starflow.__version__}"

    @classmethod
    def _validate_package_version(cls):
        """
        Check if the version number is in the correct format.
        """
        pattern = r"^starflow-py-\d+\.\d+\.\d+$"
        matched = re.match(pattern, cls.new_tag)
        if matched:
            return cls.new_tag
        raise Exception(f"Invalid tag format: {cls.new_tag}")

    @classmethod
    def _check_github_releases_versions(cls):
        """
        Check if the version already exists or if the version number is smaller than the last release.
        """
        releases = cls.github_client.get_releases(repo_name=cls.repo_name)
        if not releases:
            return
        releases_py = [release for release in releases if "starflow-py-" in release.tag_name]
        for release in releases_py:
            if release.tag_name == cls.new_tag:
                raise Exception("Version already exists.")
            release_tag_version = release.tag_name.split("-")[-1]
            if version.parse(release_tag_version) > version.parse(starflow.__version__):
                raise Exception("Version number is smaller than the last release.")
        return

    @classmethod
    def create_github_release(cls):
        """
        Create a release on github with the version from version.py file.
        """
        cls._validate_package_version()
        cls._check_github_releases_versions()

        commits = cls.github_client.get_commits(repo_name=cls.repo_name)
        last_commit = commits[0]
        last_commit_sha = last_commit.sha
        release = cls.github_client.create_release(
            repo_name=cls.repo_name,
            version=cls.new_tag,
            tag_message=f"Release {cls.new_tag} Python package",
            release_message=f"Release {cls.new_tag} Python package",
            target_commit_sha=last_commit_sha
        )
        return release


if __name__ == '__main__':
    getattr(Actions, sys.argv[1])()


import logging

LOG_LEVEL = logging.INFO
LOG_FORMAT = '%(levelname)-5s - [%(threadName)15s]' \
             ' %(name)-15s : %(lineno)d : %(asctime)-15s.%(msecs)d %(message)s'
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
logging.basicConfig(
    level=LOG_LEVEL, 
    format=LOG_FORMAT,
    datefmt=LOG_DATE_FORMAT
)
BASE_LOGGER_NAME = 'starflow-'


def override_basic_config():
    logger = logging.getLogger(BASE_LOGGER_NAME)
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    logger.setLevel(LOG_LEVEL)
    logger_handler = logging.StreamHandler()
    logger.addHandler(logger_handler)
    logger_handler.setFormatter(logging.Formatter(LOG_FORMAT, LOG_DATE_FORMAT))
    logger.propagate = False


override_basic_config()


def get_configured_logger(name: str) -> logging.Logger:
    return logging.getLogger(f'{BASE_LOGGER_NAME}.{name}')


"""
Example usage of dry run:

from starflow.scripts.piece_dry_run import piece_dry_run

piece_dry_run(
    repository_folder_path=".", 
    piece_name="DownloadPicsumImageOperator", 
    piece_input={
        "width_pixels": 200,
        "height_pixels": 200,
        "save_as_file": False
    }
)
"""
from pathlib import Path
import importlib
import json
import sys




def piece_dry_run(
    repository_folder_path: str, 
    piece_name: str, 
    piece_input: dict,
    secrets_input: dict = None,
    results_path: str = None
):
    if not results_path:
        results_path = Path('./dry_run_results')
        results_path.mkdir(parents=True, exist_ok=True)

    pieces_folder_path = str(Path(repository_folder_path).resolve() / "pieces")
    if pieces_folder_path not in sys.path:
        sys.path.append(pieces_folder_path)

    # Load Operator class
    importlib.invalidate_caches()
    piece_module = importlib.import_module(f"{piece_name}.piece")
    piece_class = getattr(piece_module, piece_name)

    # Load metadata
    metadata_path = Path(pieces_folder_path) / f"{piece_name}/metadata.json"
    with open(metadata_path, "r") as f:
        metadata = json.load(f)
    
    # Set Operator class metadata
    piece_class.set_metadata(metadata=metadata)

    # Load Operator Models
    importlib.invalidate_caches()
    piece_model_module = importlib.import_module(f"{piece_name}.models")
    piece_input_model_class = getattr(piece_model_module, "InputModel")
    piece_output_model_class = getattr(piece_model_module, "OutputModel")
    piece_secrets_model_class = getattr(piece_model_module, "SecretsModel", None)

    # Dry run Operator
    return piece_class.dry_run(
        piece_input=piece_input,
        piece_input_model=piece_input_model_class,
        piece_output_model=piece_output_model_class,
        piece_secrets_model=piece_secrets_model_class,
        secrets_input=secrets_input,
        results_path=str(results_path)
    )
